/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.runtime.lib;

import java.io.File;
import java.io.FileInputStream;
import java.io.OutputStreamWriter;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.html.HtmlPrintWriter;
import dk.vitality.lib.foundation.misc.StringUtils;

public abstract class TemplateBase
{
	@Nullable public String stringValue;
	public int intValue;
	public boolean boolValue;

	@NotNull public final String javaSourceMD5;

	protected TemplateBase(@NotNull String javaSourceMD5)
	{
		this.javaSourceMD5 = javaSourceMD5;
	}

	private static final boolean debugEmit = false;

	public void emitToStdout() throws Exception
	{
		HtmlPrintWriter pw = new HtmlPrintWriter(new OutputStreamWriter(System.out));
		emit(pw);
		pw.flush();
	}

	public void emitToStderr() throws Exception
	{
		HtmlPrintWriter pw = new HtmlPrintWriter(new OutputStreamWriter(System.err));
		emit(pw);
		pw.flush();
	}

	public void emit(@NotNull HtmlPrintWriter pw) throws Exception
	{
		if (debugEmit)
			System.err.println("Emits template: " + getClass().getSimpleName());
		TemplateData data = getData();
		if (!data.sourceMD5.equals(javaSourceMD5)) {
			throw new TemplateJavaMD5Exception("Mismatch between javasource MD5 = " + javaSourceMD5
					+ " and TemplateData MD5 = " + data.sourceMD5);
		}
		final int[] opcodes = data.opcodes;
		int pc = 0;
		while (true) {
			int encoded = opcodes[pc];
			Opcode opcode = Opcode.decodeOpcode(encoded);
			int param = Opcode.decodeParam(encoded);
			if (debugEmit)
				System.err.println("TmplOP @" + pc + ": " + Opcode.toString(encoded));
			pc++;
			switch (opcode) {
				case PRINT_STRINGTABLE:
					pw.print(data.strings[param]);
					continue;

				case PRINT_STRING_VAR:
					if (stringValue != null)
						pw.print(stringValue);
					continue;

				case PRINT_HTML_ENCODED_STRING_VAR:
					pw.htmlEncode(stringValue);
					continue;

				case PRINT_ATTR_ENCODED_STRING_VAR:
					pw.attrEncode(stringValue);
					continue;

				case PRINT_INT_VAR:
					pw.print(intValue);
					continue;

				case CALLBACK:
					executeCallback(param);
					continue;

				case GOTO:
					pc = param;
					continue;

				case CONDITIONAL_GOTO:
					if (boolValue)
						pc = param;
					continue;

				case NEGATE:
					boolValue ^= true;
					continue;

				case RETURN:
					return;

				case TEST_IS_STRING_NULL:
					boolValue = (stringValue == null);
					break;

				default:
					throw new TemplateException("Invalid opcode = " + encoded + " at addr = " + pc);
			}
		}
	}

	public static void disasm(@NotNull String path) throws Exception
	{
		File f = new File(path);
		FileInputStream fis = new FileInputStream(f);
		TemplateData data = new TemplateData(fis);
		fis.close();
		final int[] opcodes = data.opcodes;
		int pc = 0;
		while (true) {
			int encoded = opcodes[pc];
			Opcode opcode = Opcode.decodeOpcode(encoded);
			int param = Opcode.decodeParam(encoded);
			String prefix = "TmplOP @" + pc + ": " + Opcode.toString(encoded);
			switch (opcode) {
				case PRINT_STRINGTABLE -> System.err.println(prefix + " = {{{" +
				                                             StringUtils.replaceAll(data.strings[param], "\n",
						                                             "\\n") + "}}}");
				case PRINT_STRING_VAR -> System.err.println(prefix);
				case PRINT_HTML_ENCODED_STRING_VAR -> System.err.println(prefix);
				case PRINT_ATTR_ENCODED_STRING_VAR -> System.err.println(prefix);
				case PRINT_INT_VAR -> System.err.println(prefix);
				case CALLBACK -> System.err.println(prefix);
				case GOTO -> System.err.println(prefix);
				case CONDITIONAL_GOTO -> System.err.println(prefix);
				case NEGATE -> System.err.println(prefix);
				case RETURN -> {
					System.err.println(prefix);
					return;
				}
				case TEST_IS_STRING_NULL -> System.err.println(prefix);
				default -> throw new TemplateException("Invalid opcode = " + encoded + " at addr = " + pc);
			}
			pc++;
		}
	}

	protected void executeCallback(int idx) throws Exception
	{
		throw new TemplateException("Template has been disabled by changing the generated source");
	}

	@NotNull
	public TemplateData getData() throws Exception
	{
		return TemplateDataLoader.load(getClass());
	}
}
