/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.buildsystem;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.jabhts.translator.compiler.AnnotationCompiler;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.FileUtils;
import org.jabhts.translator.globals.CompilerOptions;
import org.jabhts.translator.globals.PersistentCompilationInfo;
import org.jabhts.translator.globals.ProgressInfo;

public class Builder
{
	protected CommandLineOptionsDecoder optionsDecoder;
	public CompilerOptions options;
	public File scanBaseDir;
	public String scanBaseDottedPackageName;
	public ProgressInfo progressInfo;


	public static void main(String[] args) throws Exception
	{
		new Builder().run(args);
	}

	public void run(String[] args) throws Exception
	{
		//noinspection UnusedCatchParameter
		try {
			decodeOptions(args);
			postDecodeOptions();
			progressInfo = new ProgressInfo(options);
			if (options.verbose) {
				System.err.println("  source-output = " + options.generatedSourceBaseDirF);
				System.err.println("  scanBaseDir   = " + scanBaseDir);
				System.err.println("  scan-pkg      = " + scanBaseDottedPackageName);
			}
			if (!scanBaseDir.exists())
				throw new IOException("Directory does not exist: " + scanBaseDir.getPath());
			try {
				performScanning();
			}
			finally {
				try {
					options.saveSourceInfo();
				}
				catch (Exception ex) {
					System.err.println("Error saving SourceInfo: " + CompilerUtils.getSomeMessage(ex));
					ex.printStackTrace();
				}
			}
		}
		catch (InvalidCommandLineOptionException ex) {
			optionsDecoder.usage(ex.getMessage());
		}
		catch (Exception ex) {
			System.err.println("Got exception: " + CompilerUtils.getSomeMessage(ex));
			ex.printStackTrace();
		}
	}

	public void makeCommandLineOptionsDecoder()
	{
		optionsDecoder = new CommandLineOptionsDecoder();
	}

	public void decodeOptions(String[] args) throws Exception
	{
		makeCommandLineOptionsDecoder();
		optionsDecoder.main(args);
		options = optionsDecoder.options;
		scanBaseDir = optionsDecoder.scanBaseDir;
		scanBaseDottedPackageName = optionsDecoder.scanBaseDottedPackageName;
	}

	protected void postDecodeOptions() throws Exception
	{
		if (optionsDecoder.clean) {
			System.err.println("## Cleans jannot temps.");
			if (options.sourceInfoF.exists() && !options.sourceInfoF.delete())
				throw new IOException("Can't delete " + options.sourceInfoF.getPath());
			if (options.generatedTemplateDataBaseDirF.exists()) {
				String[] cmd = { "rm", "-r", options.generatedTemplateDataBaseDirF.getPath() };
				Process p = Runtime.getRuntime().exec(cmd);
				p.waitFor();
				if (p.exitValue() != 0)
					throw new IOException("Can't delete " + options.generatedTemplateDataBaseDirF.getPath());
			}
		}
	}

	public void performScanning() throws Exception
	{
		System.err.println("Compiles templates.");
		options.loadSourceInfo();
		BuilderScanner scanner = makeBuilderScanner();
		if (options.checkForNewFiles)
			scanner.makeFullScanning();
		else
			scanner.scanKnownFiles();
	}

	protected BuilderScanner makeBuilderScanner()
	{
		return new BuilderScanner();
	}

	protected class BuilderScanner extends SourceScanner
	{
		private int directoryCount;
		private Set<File> seenFiles;

		void scanKnownFiles() throws Exception
		{
			List<File> list = new ArrayList<>(options.sourceInfoMap.keySet());
			// protect against ConcurrentModificationException
			for (File f : list) {
				if (!f.exists()) {
					System.err.println("Missing known file: " + f);
					continue;
				}
				SourceInfo si = options.sourceInfoMap.get(f);
				processFile(f.getName(), f, si.dottedPackageName);
			}
			progressInfo.printResult();
		}

		void makeFullScanning() throws Exception
		{
			seenFiles = new HashSet<>();
			scanDirectory(scanBaseDir, scanBaseDottedPackageName);
			progressInfo.printResult();
			Set<File> unseen = new HashSet<>(options.sourceInfoMap.keySet());
			unseen.removeAll(seenFiles);
			for (File f : unseen) {
				System.err.println("File no longer exists: " + f);
				options.sourceInfoMap.remove(f);
			}
		}

		public void scanDirectory(File dirF, String dottedPackageName) throws Exception
		{
			//System.err.println("compileSubDir: " + dirF.getPath());
			if (++directoryCount >= 10) {
				progressInfo.scannedDirectory();
				directoryCount = 0;
			}
			super.scanDirectory(dirF, dottedPackageName);
		}

		protected void processFile(String filename, File f, String dottedPackageName) throws Exception
		{
			if (!filename.endsWith(".java") || filename.equals("module-info.java"))
				return;
			String bn = FileUtils.basename(filename);
			if (!CompilerUtils.isValidJavaIdentifier(bn)) {
				System.err.println("Ignores invalid named java source: " + f.getPath());
				return;
			}
			if (seenFiles != null)
				seenFiles.add(f);
			FileCompiler mc = makeFileCompiler(f, dottedPackageName);
			mc.compileFile();
		}
	}

	protected FileCompiler makeFileCompiler(File javaF, String dottedPackageName)
	{
		return new FileCompiler(progressInfo, javaF, dottedPackageName, options) {
			protected AnnotationCompiler makeAnnotationCompiler(PersistentCompilationInfo pci, CompilationInfo compilationInfo)
			{
				return new AnnotationCompiler(options, pci, javaFI, loadedJavaClass, log, compilationInfo);
			}
		};
	}
}
