/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.application.attribute.content;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import org.jabhts.translator.compiler.application.attribute.AttributeEmitter;
import org.jabhts.translator.compiler.application.baseclass.AbstractBooleanApplication;
import org.jabhts.translator.compiler.application.scope.TargetingInfo;
import org.jabhts.translator.compiler.codeGeneration.api.PatchedGoto;
import org.jabhts.translator.compiler.tree.ElementEmitters;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Element;

public class BooleanAttributeApplication extends AbstractBooleanApplication
{
	public final String attrName;
	public final boolean inverse;

	public BooleanAttributeApplication(TargetingInfo target,
	                                   String attrName, boolean inverse) throws CompilerException
	{
		super(target);
		this.attrName = attrName;
		this.inverse = inverse;
	}

	public void createEmitter(ElementEmitters emitters, Element e) throws CompilerException
	{
		BooleanAttributeEmitter emitter = (BooleanAttributeEmitter)emitters.eAttributes.get(attrName);
		if (emitter != null) {
			if (emitter.extraApps == null)
				emitter.extraApps = new ArrayList<>(1);
			emitter.extraApps.add(this);
		} else {
			emitter = new BooleanAttributeEmitter(emitters, this, attrName);
			emitters.eAttributes.put(attrName, emitter);
		}
	}

	/*
	@Selected  (not inverse)
		callback(1)
		if false -> goto post
		callback(2)
		if false -> goto post
		callback(3)
		if false -> goto post
		emit ' selected'
	post:


	@Enabled (inverse)
		callback(1)
		if false -> goto emit
		callback(2)
		if false -> goto emit
		callback(3)
		if true -> goto post
	emit:
		emit ' disabled'
	post:

	 */
	private static class BooleanAttributeEmitter extends AttributeEmitter<BooleanAttributeApplication>
	{
		@Nullable List<BooleanAttributeApplication> extraApps;

		private BooleanAttributeEmitter(ElementEmitters emitters, BooleanAttributeApplication handler, String... lcAttributeNames)
		{
			super(emitters, handler, lcAttributeNames);
		}

		public void generateAttribute() throws CompilerException
		{
			List<PatchedGoto> gotos = emitExtraTests();
			//
			handler.emitBooleanCallback(program);
			if (!handler.inverse)
				program.addNegateBoolean();
			PatchedGoto pg0 = program.addPatchedConditionalGoto();
			//
			for (int i = 0; i < gotos.size(); i++)
				if (extraApps.get(i).inverse)
					gotos.get(i).patchToCurrentAddress();
			//
			program.addEmitChar(' ');
			program.addEmitString(lcAttributeNames[0]);
			pg0.patchToCurrentAddress();
			for (int i = 0; i < gotos.size(); i++)
				if (!extraApps.get(i).inverse)
					gotos.get(i).patchToCurrentAddress();
		}

		private List<PatchedGoto> emitExtraTests() throws CompilerException
		{
			if (extraApps == null)
				return Collections.emptyList();
			List<PatchedGoto> gotos = new ArrayList<>(extraApps.size());
			for (BooleanAttributeApplication app : extraApps) {
				app.emitBooleanCallback(program);
				program.addNegateBoolean();
				gotos.add(program.addPatchedConditionalGoto());
			}
			return gotos;
		}
	}
}
