/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.application.scope;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;
import org.jabhts.annotation.flowControl.EScopeChilds;
import org.jabhts.annotation.flowControl.EScopeTag;
import org.jabhts.annotation.global.NotSpecified;
import org.jabhts.annotation.global.TCompile;
import org.jabhts.annotation.global.TIgnore;
import org.jabhts.annotation.global.TOverrides;
import org.jabhts.translator.buildsystem.CompilationInfo;
import org.jabhts.translator.compiler.application.baseclass.Application;
import org.jabhts.translator.compiler.application.baseclass.InvalidApplicationException;
import org.jabhts.translator.compiler.codeGeneration.impl.ProgramEmit;
import org.jabhts.translator.compiler.utils.AnnotationHelper;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.InternalCompilerException;
import org.jabhts.translator.compiler.utils.LogLevel;
import org.w3c.dom.Element;

public class ScopeImpl extends Scope
{
	public final ScopeImpl parent;
	public final Method scopeInstanceCreatorMethod;
	public final String scopeClassname;
	public final CompilationInfo compilationInfo;

	private static final SelectorType[] selectorTypes = SelectorType.values();
	protected final Map<SelectorType, Map<String, List<Application>>> dict;
	protected final List<Application> applicationList;

	protected final Map<String, ScopeImpl> subScopeMap;
	protected final Set<String> scopeNames;

	public ScopeImpl(CompilerLog log, Class<?> scopeClass, CompilationInfo compilationInfo)
	{
		this(log, scopeClass, "root", "scopeRoot", null, compilationInfo, null);
	}

	protected ScopeImpl(CompilerLog log, Class<?> scopeClass,
	                    String scopeName, String scopeFieldName,
	                    ScopeImpl parent, CompilationInfo compilationInfo, Type optConstructorArgument)
	{
		super(log, scopeClass, scopeName, scopeFieldName);
		this.parent = parent;
		this.compilationInfo = compilationInfo;
		dict = new EnumMap<>(SelectorType.class);
		for (SelectorType type : selectorTypes)
			dict.put(type, new HashMap<>());
		applicationList = new ArrayList<>();
		subScopeMap = new HashMap<>();
		if (parent == null) {
			scopeInstanceCreatorMethod = null;
			scopeClassname = scopeClass.getSimpleName();
			scopeNames = new HashSet<>();
		} else {
			scopeNames = parent.scopeNames;
			scopeClassname = parent.scopeClassname + "." + scopeClass.getSimpleName().replace('$', '.');
			scopeInstanceCreatorMethod = getScopeInstanceCreatorMethod(optConstructorArgument);
		}
		scopeNames.add(scopeFieldName);
	}

	protected Method getScopeInstanceCreatorMethod(Type optConstructorArgument)
	{
		String methodName = "new" + CompilerUtils.getInnerclassName(scopeClass);
		String wantedArgType = (optConstructorArgument != null) ? boxedType(optConstructorArgument.toString()) : null;
		int wantedNumArguments = (optConstructorArgument != null) ? 1 : 0;
		for (Class<?> cls = parent.scopeClass; cls != null; cls = cls.getSuperclass()) {
			for (Method m : cls.getDeclaredMethods()) {
				if (!m.getName().equals(methodName))
					continue;
				Type[] actualArgs = m.getGenericParameterTypes();
				if (actualArgs.length != wantedNumArguments) {
					log.log(LogLevel.Error, "Wrong number of arguments for scope instance creating method: " + m);
					continue;
				}
				if (wantedNumArguments == 0)
					return m;
				if (actualArgs[0].equals(optConstructorArgument))
					return m;
				String actual = boxedType(actualArgs[0].toString());
				if (actual.equals(wantedArgType))
					return m;
				log.log(LogLevel.Error, "Wrong argument type for scope instance creating method: " + m);
			}
		}
		return null;
	}

	protected static String boxedType(String type)
	{
		String s = boxedType.get(type);
		return (s != null) ? s : type;
	}

	protected static final Map<String, String> boxedType = new HashMap<>();

	static {
		boxedType.put("boolean", "class java.lang.Boolean");
		boxedType.put("byte", "class java.lang.Byte");
		boxedType.put("short", "class java.lang.Short");
		boxedType.put("int", "class java.lang.Integer");
		boxedType.put("long", "class java.lang.Long");
		boxedType.put("float", "class java.lang.Float");
		boxedType.put("double", "class java.lang.Double");
	}

	public Iterable<ScopeImpl> iterSubScopes()
	{
		return subScopeMap.values();
	}

	//=====================================================================================================
	public List<ApplicationMatcher> makeMatchersForElement(Element e) throws CompilerException
	{
		List<ApplicationMatcher> list = new ArrayList<>();
		//
		String id = e.getAttribute("id");
		if (id != null && !id.isEmpty())
			list.add(new ApplicationMatcher(e, SelectorType.Id, makeJavaKey(id), id));
		//
		for (String s : CompilerUtils.splitCssClassNames(e.getAttribute("class")))
			list.add(new ApplicationMatcher(e, SelectorType.Class, makeJavaKey(s), s));
		//
		list.add(new ApplicationMatcher(e, SelectorType.Tag, makeJavaKey(e.getTagName()), e.getTagName()));
		//
		return list;
	}


	public List<Application> getApplicationsForElement(List<ApplicationMatcher> matchers, boolean addParents)
			throws CompilerException
	{
		List<Application> list = new ArrayList<>();
		for (ApplicationMatcher matcher : matchers) {
			matcher.matched |= doGetApplications(list, matcher.type, matcher.javaKey, addParents, null);
		}
		return list;
	}

	public void makeNonMatchedWarnings(List<ApplicationMatcher> matchers) throws CompilerException
	{
		for (ApplicationMatcher matcher : matchers) {
			if (matcher.matched)
				continue;
			if (matcher.origKey.charAt(0) == '_')
				log.log(LogLevel.Notice,
						matcher.type.description + "=" + matcher.origKey + " not referred by annotation");//, java name = " + javaKey);
		}
	}

	protected String makeJavaKey(String key)
	{
		if (key.indexOf('_') < 0 && key.indexOf('-') < 0 && key.indexOf('.') < 0)
			return key.toLowerCase();
		int len = key.length();
		StringBuilder sb = new StringBuilder(len);
		for (int i = 0; i < len; i++) {
			char ch = key.charAt(i);
			if (ch == '_' || ch == '-' || ch == '.')
				continue;
			sb.append(Character.toLowerCase(ch));
		}
		return sb.toString();
	}

	protected boolean doGetApplications(List<Application> list, SelectorType type, String javaKey, boolean addParents,
	                                    Set<String> maskedOverrides)
	{
		List<Application> matched = dict.get(type).get(javaKey);
		boolean result = false;
		if (matched != null) {
			if (maskedOverrides == null) {
				for (Application processor : matched) {
					processor.notifyMatched(type);
					list.add(processor);
					result = true;
					if (!processor.target.hasTOverrides)
						continue;
					log.log(LogLevel.Verbose, "hasTOverrides: " + processor.annotationToString());
					if (maskedOverrides == null)
						maskedOverrides = new HashSet<>();
					maskedOverrides.add(processor.overrideKey());
				}
			} else {
				for (Application processor : matched) {
					if (maskedOverrides.contains(processor.overrideKey())) {
						log.log(LogLevel.Verbose, "Ignored " + processor.annotationToString());
						continue;
					}
					processor.notifyMatched(type);
					list.add(processor);
					result = true;
				}
				for (Application processor : matched) {
					if (processor.target.hasTOverrides)
						maskedOverrides.add(processor.overrideKey());
				}
			}
		}
		if (addParents && parent != null)
			result |= parent.doGetApplications(list, type, javaKey, addParents, maskedOverrides);
		return result;
	}

	public void emitPostCompilationWarnings() throws CompilerException
	{
		for (Map<String, List<Application>> map : dict.values()) {
			for (List<Application> list : map.values()) {
				for (Application appl : list)
					appl.emitPostCompilationWarnings();
			}
		}
		for (ScopeImpl scope : subScopeMap.values())
			scope.emitPostCompilationWarnings();
	}
	//=====================================================================================================

	private ScopeImpl getSubScope(Class<?> inner, Type optConstructorArgument) throws CompilerException
	{
		//log.log(LogLevel.Notice, "GetSubScope: " + inner.getSimpleName());
		for (ScopeImpl scope = this; scope != null; scope = scope.parent) {
			if (scope.hasProperEnclosingClass(inner))
				return scope.getOrCreateSubScope(inner, optConstructorArgument);
		}
		log.log(LogLevel.Error, "Scoping inner-class (" + inner.getSimpleName() + ")"
				+ " is not enclosed by parent scope (" + scopeClass.getSimpleName() + ")"
				+ " but by " + inner.getEnclosingClass().getSimpleName());
		return null;
	}

	private ScopeImpl getOrCreateSubScope(Class<?> inner, Type optConstructorArgument) throws CompilerException
	{
		String key = CompilerUtils.getInnerclassBasename(inner);
		ScopeImpl scope = subScopeMap.get(key);
		if (scope != null)
			return scope;
		String name = key;
		for (int i = 2; scopeNames.contains(name); i++)
			name = key + i;
		scope = new ScopeImpl(log, inner, name, "scope" + name, this, compilationInfo, optConstructorArgument);
		subScopeMap.put(key, scope);
		scope.process();
		return scope;
	}

	public ScopeImpl getSubScope(Class<?> inner, TargetingInfo target, Type optConstructorArgument)
			throws CompilerException
	{
		if (inner != null && inner != NotSpecified.class) {
			if (inner.getEnclosingClass() == null)
				return getOrCreateSubScope(inner, optConstructorArgument);
			return getSubScope(inner, optConstructorArgument);
		}
		//log.log(LogLevel.Notice, "GetSubScope: " + target.annotationToString());
		for (Class<?> cls = scopeClass; cls != null; cls = cls.getSuperclass()) {
			for (Class<?> i2 : cls.getDeclaredClasses()) {
				String bn = CompilerUtils.getInnerclassBasename(i2);
				//log.log(LogLevel.Notice, "SubScope candidate: " + bn + "   for " + target.basename);
				if (bn.equals(target.basename))
					return getSubScope(i2, optConstructorArgument);
			}
		}
		log.log(LogLevel.Error, "Can't locate innerclass for scope '" + target.basename
				+ "' inside " + scopeClassname + ": " + target.annotationToString());
		if (1 == 2) {
			for (Class<?> cls = scopeClass; cls != null; cls = cls.getSuperclass()) {
				log.log(LogLevel.Notice, "Checks SubScope candidates in: " + cls.getSimpleName());
				for (Class<?> i2 : cls.getDeclaredClasses()) {
					String bn = CompilerUtils.getInnerclassBasename(i2);
					log.log(LogLevel.Notice, "SubScope candidate: " + bn + "   for " + target.basename
							+ " in " + cls.getSimpleName());
					if (bn.trim().isEmpty())
						log.log(LogLevel.Notice,
								"  can't make innerclass name: " + i2.getName() + " -> '" + i2.getSimpleName() + "'");
				}
			}
		}
		return null;
	}

	private boolean hasProperEnclosingClass(Class<?> inner)
	{
		Class<?> enclosing = inner.getEnclosingClass();
		for (Class<?> cls = scopeClass; cls != null; cls = cls.getSuperclass()) {
			if (cls == enclosing)
				return true;
		}
		return false;
	}

	//=====================================================================================================

	public void process() throws CompilerException
	{
		Map<String, DelayedAnnotationHelper> seen = new HashMap<>();
		processClass(scopeClass, seen, false);
		for (DelayedAnnotationHelper dah : seen.values())
			dah.emit();
		log.log(LogLevel.Verbose,
				"ScopeImpl." + scopeFieldName + " handler count: id = " + dict.get(SelectorType.Id).size()
						+ ", class = " + dict.get(SelectorType.Class).size()
						+ ", tag = " + dict.get(SelectorType.Tag).size()
						+ ", subs = " + subScopeMap.size());
		if (parent == null) {
			// processing innerclasses is delayed as to have the full declaration of the parent scope
			// before processing the subscope as we otherwise can't make proper "overrides" warnings.
			processInnerClasses(scopeClass, false);
		}
		for (Application appl : applicationList)
			appl.makeSubScope(this);
	}

	protected void processClass(Class<?> cls,
	                            Map<String, DelayedAnnotationHelper> seen,
	                            boolean isInherited) throws CompilerException
	{
		if (cls == null || cls.getName().startsWith("java."))
			return; // ignore java.lang.Object
		if (!cls.isAnnotationPresent(TIgnore.class)) {
			if (parent == null) {
				for (Annotation a : cls.getDeclaredAnnotations())
					processAnnotation(new AnnotationHelper(log, a, cls, null, null), isInherited);
			}
			for (Field f : cls.getDeclaredFields()) {
				if (f.isAnnotationPresent(TIgnore.class))
					continue;
				Annotation[] annotations = f.getAnnotations();
				if (annotations.length == 0)
					continue;
				String key = f.getType().getName() + " " + f.getName();
				DelayedAnnotationHelper dah = seen.get(key);
				if (dah != null) {
					if (dah.annotations != null)
						dah.checkSuperAnnotation("field " + key, f, null, annotations, cls);
					continue;
				}
				dah = new DelayedAnnotationHelper(cls, f, null, annotations, isInherited);
				seen.put(key, dah);
			}
			for (Method m : cls.getDeclaredMethods()) {
				if (m.isAnnotationPresent(TIgnore.class))
					continue;
				Annotation[] annotations = m.getAnnotations();
				if (annotations.length == 0)
					continue;
				String key = m.getReturnType().getName() + " " + m.getName() + " " +
						Arrays.toString(m.getParameterTypes());
				DelayedAnnotationHelper dah = seen.get(key);
				if (dah != null) {
					if (dah.annotations != null)
						dah.checkSuperAnnotation("method " + key, null, m, annotations, cls);
					continue;
				}
				dah = new DelayedAnnotationHelper(cls, null, m, annotations, isInherited);
				seen.put(key, dah);
			}
		}
		processClass(cls.getSuperclass(), seen, true);
	}

	/**
	 * Javac 1.6 generates hidden overriden methods for some methods in the superclass.
	 * I have no idea why it does it, or when; it seems to not always do it.
	 * This delayed helper uses the annotations on the method in the superclass when the subclass' method
	 * doesn't have any annotations, i.e. the generated stub method is ignored.
	 */
	class DelayedAnnotationHelper
	{
		Class<?> cls;
		Field field;
		Method method;
		Annotation[] annotations;
		final boolean isInherited;

		DelayedAnnotationHelper(Class<?> cls, Field field, Method method, Annotation[] annotations, boolean isInherited)
		{
			this.cls = cls;
			this.field = field;
			this.method = method;
			this.annotations = annotations;
			this.isInherited = isInherited;
		}

		void checkSuperAnnotation(String what, Field field, Method method, Annotation[] a2, Class<?> ocls)
		{
			if (annotations.length == 0 && a2.length > 0) {
				this.field = field;
				this.method = method;
				this.annotations = a2;
				this.cls = ocls;
				return;
			}
			String s1 = toString(annotations);
			String s2 = toString(a2);
			s1 = normalizeAnnotArgRE.matcher(s1).replaceAll("");
			s2 = normalizeAnnotArgRE.matcher(s2).replaceAll("");
			if (s1.equals(s2))
				return;
			what = ProgramEmit.javaLangRE.matcher(what).replaceAll("");
			s1 = annotDotClassRE.matcher(s1).replaceAll("@"); // @org.jabhts.annotation.flowControl.EIterate
			s2 = annotDotClassRE.matcher(s2).replaceAll("@"); // -> @EIterate
			s1 = defaultAnnotArg1RE.matcher(s1).replaceAll("");
			s2 = defaultAnnotArg1RE.matcher(s2).replaceAll("");
			s1 = defaultAnnotArg2RE.matcher(s1).replaceAll(")");
			s2 = defaultAnnotArg2RE.matcher(s2).replaceAll(")");
			log.log(LogLevel.Notice, "Overrides using different annotations: " + what + "\n" +
					"  " + cls.getSimpleName() + " uses " + s1 + "\n" +
					"  " + ocls.getSimpleName() + " uses " + s2);
		}

		String toString(Annotation[] a)
		{
			String[] ss = new String[a.length];
			for (int i = 0; i < a.length; i++)
				ss[i] = a[i].toString();
			Arrays.sort(ss);
			return Arrays.toString(ss);
		}

		void emit() throws CompilerException
		{
			for (Annotation a : annotations)
				processAnnotation(new AnnotationHelper(log, a, cls, field, method), isInherited);
		}
	}

	public static final Pattern normalizeAnnotArgRE = Pattern.compile("\\bwarnIfUnused=(?:true|false)(,\\s*)?");
	public static final Pattern annotDotClassRE = Pattern.compile("@([a-z]\\w+\\.)+");
	public static final Pattern defaultAnnotArg1RE = Pattern.compile("\\b[a-z]\\w*=\\[\\](,\\s*)?");
	public static final Pattern defaultAnnotArg2RE = Pattern.compile(", \\)");

	protected void processInnerClasses(Class<?> cls, boolean isInherited) throws CompilerException
	{
		if (cls == null || cls.getName().startsWith("java."))
			return; // ignore java.lang.Object
		processInnerClasses(cls.getSuperclass(), isInherited);
		if (cls.isAnnotationPresent(TIgnore.class))
			return;
		for (Class<?> inner : cls.getDeclaredClasses()) {
			//log.log(LogLevel.Verbose, "Inner class " + inner.getSimpleName() + " inside " + cls.getSimpleName());
			EScopeChilds scopeChilds = inner.getAnnotation(EScopeChilds.class);
			if (scopeChilds != null) {
				log.log(LogLevel.Verbose, "Inner class " + inner.getSimpleName() + " has @EScopeChilds: " + Modifier
						.toString(inner.getModifiers()));
				getSubScope(inner, null);
				processAnnotation(new AnnotationHelper(log, scopeChilds, inner, null, null), isInherited);
			}
			EScopeTag scopeTag = inner.getAnnotation(EScopeTag.class);
			if (scopeTag != null) {
				log.log(LogLevel.Verbose, "Inner class " + inner.getSimpleName() + " has @EScopeTag: " + Modifier
						.toString(inner.getModifiers()));
				getSubScope(inner, null);
				processAnnotation(new AnnotationHelper(log, scopeTag, inner, null, null), isInherited);
			}
		}
	}

	protected void processAnnotation(AnnotationHelper helper, boolean isInherited) throws CompilerException
	{
		ApplicationFactory factory = AnnotationApplicationMap.get(helper.annotation);
		if (factory == null) {
			String annotationClassname = helper.annotation.annotationType().getName();
			if (annotationClassname.startsWith("org.jabhts.annotation.")
					&& !(helper.annotation instanceof TCompile)
					&& !(helper.annotation instanceof TOverrides)) {
				log.log(LogLevel.Notice, "Missing handler for annotation: " + annotationClassname);
			}
			return;
		}
		TargetingInfo target;
		Application appl;
		try {
			target = new TargetingInfo(this, helper, isInherited);
			appl = factory.create(this, target);
		}
		catch (InvalidApplicationException ex) {
			log.log(LogLevel.Error, CompilerUtils.getSomeMessage(ex));
			return;
		}
		applicationList.add(appl);
		switch (target.type) {
			case DefaultParam:
				addApplication(SelectorType.Id, target.defparam, appl);
				addApplication(SelectorType.Class, target.defparam, appl);
				break;
			case Triplets:
				addApplication(SelectorType.Id, target.id, appl);
				addApplication(SelectorType.Class, target.className, appl);
				addApplication(SelectorType.Tag, target.tag, appl);
				break;
			case Appliance:
				addApplication(SelectorType.Id, target.basename, appl);
				addApplication(SelectorType.Class, target.basename, appl);
				break;
			default:
				throw new InternalCompilerException("Invalid value of target.type = " + target.type);
		}
	}

	protected void addApplication(SelectorType type, String[] keys, Application processor) throws CompilerException
	{
		for (String key : keys)
			addApplication(type, key, processor);
	}

	protected void addApplication(SelectorType type, String key, Application processor
	) throws CompilerException
	{
		if (key == null || key.isEmpty())
			throw new InternalCompilerException("key is null/empty-string");
		key = key.toLowerCase();
		Map<String, List<Application>> map = dict.get(type);
		List<Application> list = map.get(key);
		if (list == null) {
			list = new ArrayList<>();
			map.put(key, list);
		}
		list.add(processor);
	}
}
