/* Copyright (c) 2007-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.htmlLoader;

import java.io.ByteArrayInputStream;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

public class XHtmlDomParser implements ErrorHandler
{
	protected DocumentBuilder parser;
	protected Exception fatalErrorException;

	protected XHtmlDomParser(DocumentBuilder parser)
	{
		this.parser = parser;
		parser.setErrorHandler(this);
	}

	public void warning(SAXParseException ex)
	{
		System.err.println("[Warning] " + getLocationString(ex) + ": " + ex.getMessage());
	}

	public void error(SAXParseException ex)
	{
		System.err.println("[Error] " + getLocationString(ex) + ": " + ex.getMessage());
		if (fatalErrorException == null)
			fatalErrorException = ex;
	}

	public void fatalError(SAXParseException ex) throws SAXException
	{
		System.err.println("[Fatal Error] " + getLocationString(ex) + ": " + ex.getMessage());
		throw ex;
	}

	public Document parse(byte[] b) throws Exception
	{
		Document doc = parser.parse(new InputSource(new ByteArrayInputStream(b)));
		if (fatalErrorException != null)
			throw fatalErrorException;
		return doc;
	}

	private String getLocationString(SAXParseException ex)
	{
		StringBuilder sb = new StringBuilder();
		String systemId = ex.getSystemId();
		if (systemId != null) {
			int index = systemId.lastIndexOf('/');
			if (index != -1)
				systemId = systemId.substring(index + 1);
			sb.append(systemId);
		}
		sb.append(':');
		sb.append(ex.getLineNumber());
		sb.append(':');
		sb.append(ex.getColumnNumber());
		return sb.toString();
	}

	static XHtmlDomParser makeSimpleParser() throws ParserConfigurationException
	{
		if (factory == null) {
			factory = DocumentBuilderFactory.newInstance();
			factory.setExpandEntityReferences(true);
			factory.setIgnoringComments(false);
			factory.setIgnoringElementContentWhitespace(false);
			factory.setNamespaceAware(true);
			factory.setValidating(false);
			factory.setXIncludeAware(false);
			setAttribute("http://apache.org/xml/features/dom/defer-node-expansion", false);
			setAttribute("http://apache.org/xml/features/validation/schema", false);
			setAttribute("http://apache.org/xml/features/validation/schema-full-checking", false);
			setAttribute("http://apache.org/xml/features/warn-on-duplicate-entitydef", true);
			setAttribute("http://apache.org/xml/features/validation/warn-on-duplicate-attdef", true);
			setAttribute("http://apache.org/xml/features/standard-uri-conformant", true);
		}
		return new XHtmlDomParser(factory.newDocumentBuilder());
	}

	@SuppressWarnings({"UnusedCatchParameter"})
	private static void setAttribute(String url, Object o)
	{
		try {
			factory.setAttribute(url, o);
		}
		catch (Exception ex) {
			// ignore as different version of the parser has different urls.
			System.err.println("Parser attr unknown: " + url);
		}
	}

	private static DocumentBuilderFactory factory;
}
