/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.tree;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.jabhts.translator.compiler.application.attribute.AttributeEmitter;
import org.jabhts.translator.compiler.application.attribute.PostAttributeEmitter;
import org.jabhts.translator.compiler.application.attribute.sub.SubAttributeEmitter;
import org.jabhts.translator.compiler.application.baseclass.Application;
import org.jabhts.translator.compiler.application.element.ElementLoopEmitter;
import org.jabhts.translator.compiler.application.element.scoped.EScopeChildsApplication;
import org.jabhts.translator.compiler.application.element.scoped.EScopeTagApplication;
import org.jabhts.translator.compiler.application.element.simple.EContentApplication;
import org.jabhts.translator.compiler.application.element.simple.EContentIfApplication;
import org.jabhts.translator.compiler.application.element.simple.EContentPrePostApplication;
import org.jabhts.translator.compiler.application.element.simple.EIfApplication;
import org.jabhts.translator.compiler.application.element.simple.EReplaceApplication;
import org.jabhts.translator.compiler.application.element.simple.ETagIfApplication;
import org.jabhts.translator.compiler.application.element.simple.EZapApplication;
import org.jabhts.translator.compiler.application.scope.ApplicationMatcher;
import org.jabhts.translator.compiler.application.scope.Scope;
import org.jabhts.translator.compiler.codeGeneration.api.Program;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.LogLevel;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class ElementEmitters
{
	public final CompilerLog log;
	public final Program program;
	public boolean allowIdAttribute;
	public boolean shouldProcessElement;

	public final List<EIfApplication.Emitter> eConditionals = new ArrayList<>();
	public final List<ETagIfApplication.Emitter> eTag = new ArrayList<>();
	public final List<EContentIfApplication.Emitter> eContentIf = new ArrayList<>();
	public final List<EContentPrePostApplication.Emitter> ePrePostContent = new ArrayList<>();

	public ElementLoopEmitter<?> eLoop;
	public EReplaceApplication.Emitter eReplacer;
	public EScopeChildsApplication.Emitter eInnerScope;
	public EScopeTagApplication.Emitter eOuterScope;
	public EContentApplication.Emitter eContent;

	public final Map<String, AttributeEmitter<?>> eAttributes = new HashMap<>();
	public final Map<String, PostAttributeEmitter<?>> ePostAttributes = new HashMap<>();
	public final Map<String, List<SubAttributeEmitter<?>>> eSubAttributes = new HashMap<>();


	final Element e;
	final String nodeName;
	final int level;
	/* final */ boolean debugFlag;

	final ScopeLink parentScopeLink;
	ScopeLink scopeLink;

	ElementEmitters(CompilerLog log, boolean debugFlag, Program program,
	                ScopeLink scopeLink,
	                Element e, int level,
	                boolean allowIdAttribute) throws CompilerException
	{
		this.log = log;
		this.debugFlag = debugFlag;
		this.program = program;
		this.e = e;
		this.level = level;
		this.nodeName = e.getNodeName();
		if (debugFlag)
			log.log(LogLevel.Verbose, indent() + "Element: " + nodeName);
		this.shouldProcessElement = true;
		this.allowIdAttribute = allowIdAttribute;
		this.parentScopeLink = scopeLink;
		this.scopeLink = scopeLink;
		processApplications();
	}

	private void processApplications() throws CompilerException
	{
		List<ApplicationMatcher> matchers = scopeLink.scope.makeMatchersForElement(e);
		List<Application> list = scopeLink.scope.getApplicationsForElement(matchers, true);
		EScopeChildsApplication innerScopeAppl = null;
		EScopeTagApplication outerScopeAppl = null;
		for (Application appl : list) {
			if (appl instanceof EZapApplication) {
				shouldProcessElement = false;
				return;
			}
			if (appl instanceof EScopeChildsApplication) {
				if (innerScopeAppl != null) {
					log.log(LogLevel.Error, indent() + "Has 2 scoping annotations for one element: {{"
							+ innerScopeAppl.annotationToString() + "}} and {{" + appl.annotationToString() + "}}");
					shouldProcessElement = false;
					return;
				}
				innerScopeAppl = (EScopeChildsApplication)appl;
			}
			if (appl instanceof EScopeTagApplication) {
				if (outerScopeAppl != null) {
					log.log(LogLevel.Error, indent() + "Has 2 scoping annotations for one element: {{"
							+ outerScopeAppl.annotationToString() + "}} and {{" + appl.annotationToString() + "}}");
					shouldProcessElement = false;
					return;
				}
				outerScopeAppl = (EScopeTagApplication)appl;
			}
		}
		for (Application appl : list)
			appl.createEmitter(this, e);
		if (scopeLink != parentScopeLink) {
			list = scopeLink.scope.getApplicationsForElement(matchers, false);
			log.log(LogLevel.Verbose, "Nested scope: " + parentScopeLink.scope.scopeName + " -> " + scopeLink.scope.scopeName + " (" + list.size() + " subAppls)");
			for (Application appl : list) {
				if (appl instanceof EZapApplication) {
					Application scope = (innerScopeAppl != null) ? innerScopeAppl : outerScopeAppl;
					log.log(LogLevel.Error, indent() + "EZap is not allowed in nested scope, " +
							"but must be applied at same level as " + scope.annotationToString());
					shouldProcessElement = false;
					return;
				}
				appl.createEmitter(this, e);
			}
		}
		scopeLink.scope.makeNonMatchedWarnings(matchers);
	}

	public boolean alreadyDefined(AbstractEmitter<?> previous, AbstractEmitter<?> emitter, String what)
	{
		if (previous == null)
			return false;
		error("" +
				"Multiple " + what + " annotations for same element are not allowed:\n" +
				"  " + previous.annotationToString() + "\n" +
				"  " + emitter.handler.annotationToString() +
				"");
		return true;
	}

	public boolean enterSubScope(AbstractEmitter<?> emitter, Scope subScope)
	{
		if (subScope == null)
			return true;
		// Making non-used warnings become complex when getProcessors... is split over two scopes...
		//List<Application> additional = emitter.scope.getProcessorsForElement(e, false);
		//System.err.println("subscope " + usages.size() + " + " + additional.size() + ": " + usage.processor.annotationToString());
		//processors.addAll(additional);
		StringBuilder tooDeepSB = (scopeLink.level > 25) ? new StringBuilder() : null;
		ScopeLink sl = scopeLink;
		while (sl != null) {
			if (tooDeepSB != null) {
				tooDeepSB.append(".....Level ").append(sl.level).append(": ");
				tooDeepSB.append(sl.annotDescription).append("\n");
			}
			if (sl.scope == subScope) {
				error("Scopes are not reentrant:\n" +
						"......Entered " + sl.scope.scopeName + " using " + sl.annotDescription + "\n" +
						"......Reenters " + subScope.scopeName + " using " + emitter.annotationToString());
				return false;
			}
			sl = sl.prev;
		}
		if (tooDeepSB != null) {
			error("Too deep scope nesting (" + scopeLink.level + " levels!):\n" + tooDeepSB);
			return false;
		}
		scopeLink = new ScopeLink(scopeLink, subScope, emitter.annotationToString());
		return true;
	}


	protected String indent()
	{
		char[] ch = new char[level];
		Arrays.fill(ch, ' ');
		return new String(ch);
	}

	protected void error(String msg)
	{
		attrError(null, msg);
	}

	protected void attrError(@Nullable String attrName, String msg)
	{
		StringBuilder sb = new StringBuilder();
		makeLocation(sb, e);
		if (attrName != null)
			sb.append("/@").append(attrName);
		sb.append(": ").append(msg);
		log.log(LogLevel.Error, sb.toString());
	}

	/**
	 * Appends path to current node to the StringBuilder, used for error reporting.
	 */
	protected void makeLocation(StringBuilder sb, Node e)
	{
		Node parent = e.getParentNode();
		if (parent != null && parent.getNodeType() != Node.DOCUMENT_NODE) {
			makeLocation(sb, parent);
			sb.append("/");
		}
		sb.append(e.getNodeName());
		if (e.getNodeType() != Node.ELEMENT_NODE)
			sb.append("[type=").append(e.getNodeType()).append("]");
	}
}
