/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.utils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import org.jabhts.translator.globals.ProgressInfo;

public class CompilerLog implements Iterable<CompilerLog.Entry>
{
	protected List<Entry> entries = new ArrayList<>();
	protected String title, label;
	protected boolean hasAddedLabel = false;

	public CompilerLog(String title)
	{
		this.title = title;
	}

	public void setLabel(String label)
	{
		this.label = label;
		hasAddedLabel = false;
	}

	public Iterator<Entry> iterator()
	{
		return Collections.unmodifiableList(entries).iterator();
	}

	protected static class Entry
	{
		public final LogLevel level; // null for labels
		public final String msg;

		protected Entry(LogLevel level, String msg)
		{
			this.level = level;
			this.msg = msg;
		}
	}

	public void log(LogLevel level, String msg)
	{
		if (label != null && !hasAddedLabel) {
			entries.add(new Entry(null, label));
			hasAddedLabel = true;
		}
		entries.add(new Entry(level, msg));
	}

	public boolean hasError()
	{
		for (Entry e : entries) {
			if (e.level == LogLevel.Error || e.level == LogLevel.InternalError)
				return true;
		}
		return false;
	}

	public void clear()
	{
		entries.clear();
	}

	public boolean isEmpty()
	{
		return entries.isEmpty();
	}

	public String entriesToString(boolean includeVerbose)
	{
		StringBuilder sb = new StringBuilder();
		sb.append("*** ").append(title).append("\n");
		int numAdded = 0;
		String currentLabel = null;
		for (Entry entry : entries) {
			if (!includeVerbose && entry.level == LogLevel.Verbose)
				continue;
			if (entry.level == null) {
				currentLabel = entry.msg;
				continue;
			}
			if (currentLabel != null) {
				sb.append(currentLabel).append("\n");
				currentLabel = null;
			}
			numAdded++;
			sb.append("  ");
			switch (entry.level) {
				case Verbose:
					sb.append("...     ");
					break;
				case Error:
					sb.append("ERROR:  ");
					break;
				case Notice:
					sb.append("notice: ");
					break;
				default:
					sb.append(entry.level.name()).append(": ");
			}
			sb.append(entry.msg);
			sb.append("\n");
		}
		return numAdded == 0 ? null : sb.toString();
	}

	public void purgeToStderr(ProgressInfo progressInfo, boolean includeVerbose)
	{
		String s = entriesToString(includeVerbose);
		if (s != null) {
			if (progressInfo != null)
				progressInfo.flush();
			System.err.println(s.trim());
		}
		clear();
	}

}
