/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.xml.xpath;

import java.util.ArrayList;
import org.xml.sax.SAXException;

public final class PathCompiler
{
	ArrayList<PathBase> components = new ArrayList<>();
	PathBase lastPathBase = null;
	String xpath;
	int nextCharIndex, xpathLen;
	PathCompilerToken token = new PathCompilerToken();
	private static final boolean dumpCompiledPrograms = false;
	private static final boolean debugTokens = false;
	private static final boolean debugTests = false;

	public static PathProgram compile(String xpath) throws SAXException
	{
		PathCompiler pc = new PathCompiler(xpath);
		PathProgram pp = pc.getProgram();
		if (dumpCompiledPrograms) {
			System.err.println("XPath: " + xpath);
			pp.debugDump();
		}
		return pp;
	}

	protected PathCompiler(String xpath) throws SAXException
	{
		this.xpath = xpath;
		this.xpathLen = xpath.length();
		this.nextCharIndex = 0;
		try {
			compile();
		} catch (SAXException ex) {
			System.err.println("Failed to compile: {{{" + xpath + "}}}");
			throw ex;
		}
		token = null;
	}

	protected PathProgram getProgram()
	{
		PathProgram pp = new PathProgram();
		pp.paths = new PathBase[components.size()];
		components.toArray(pp.paths);
		return pp;
	}

	protected void compile() throws SAXException
	{
		boolean mustBeAtEnd = false;
		while (true) {
			nextToken(false);
			if (mustBeAtEnd && token.type != PathCompilerToken.END)
				throw new SAXException("Expected end of xpath");
			switch (token.type) {
				case PathCompilerToken.END:
					return;
				case PathCompilerToken.NAMED_CHILD:
					ChildPath namedChild = new ChildPath(token.txt);
					addComponent(namedChild);
					break;
				case PathCompilerToken.ALL_CHILDS:
					addComponent(new ChildPath(null));
					break;
				case PathCompilerToken.PARENT_ELEMENT:
					addComponent(new ParentPath());
					break;
				case PathCompilerToken.ATTRIBUTE:
					mustBeAtEnd = true;
					AttributePath pac = new AttributePath(token.txt);
					addComponent(pac);
					break;
				case PathCompilerToken.TEST_START:
					handleTest(PathCompilerToken.TEST_END);
					break;
				case PathCompilerToken.PATH_SEPARATOR:
					if (components.isEmpty())
						addComponent(new DocumentRootPath());
					break;
				case PathCompilerToken.DEEP_CHILDS:
					if (components.isEmpty())
						addComponent(new DocumentRootPath());
					addComponent(new DeepPath());
					break;
				case PathCompilerToken.CURRENT_NODE:
					addComponent(new CurrentNodePath());
					break;
				default:
					throw new SAXException("Misplaced xpath component type");
			}
		}
	}

	protected void addComponent(PathBase c)
	{
		components.add(c);
		lastPathBase = c;
	}

	protected void handleTest(int endToken) throws SAXException
	{
		if (lastPathBase == null)
			throw new SAXException("Can't have test without element");
		lastPathBase.test = getTest(endToken, true);
	}

	protected ExpressionBase getTest(int endToken, boolean isFirstTest) throws SAXException
	{
		if (debugTests)
			System.err.println("getTest.enter(" + endToken + ", " + isFirstTest + ")");
		int numItems = 0;
		ExpressionBase subTest = null;
		mainLoop:
		while (true) {
			nextToken(true);
			numItems++;
			if (token.type == endToken) {
				if (subTest == null)
					throw new SAXException("Expected non-empty expression");
				break;
			}
			switch (token.type) {
				case PathCompilerToken.END:
					throw new SAXException("At end of xpath inside test");
				case PathCompilerToken.EQUALS:
					if (subTest == null)
						throw new SAXException("Missing op to left of '='");
					nextToken(true);
					subTest = new EqualsExpression(subTest, getSubTest(endToken));
					break;
				case PathCompilerToken.NOT_EQUALS:
					if (subTest == null)
						throw new SAXException("Missing op to left of '!='");
					nextToken(true);
					subTest = new NotEqualsExpression(subTest, getSubTest(endToken));
					break;
				case PathCompilerToken.AND:
					if (subTest == null)
						throw new SAXException("Missing op to left of 'and'");
					subTest = new AndExpression(subTest, getTest(endToken, false));
					break mainLoop;

				case PathCompilerToken.INTEGER:
					if (subTest != null)
						throw new SAXException("Unexpected integer");
					if (numItems != 1)
						throw new SAXException("Unexpected integer inside test");
					if (token.integerValue <= 0)
						throw new SAXException("Invalid element index");
					if (components.isEmpty())
						throw new SAXException("Test without element component?");
					if (!(lastPathBase instanceof ChildPath))
						throw new SAXException("lastPathBase must be ChildPath");
					ChildPath pcc = (ChildPath) lastPathBase;
					pcc.wantedIndex = token.integerValue;
					nextToken(true);
					if (token.type == endToken)
						throw new SAXException("Expected end token");
					return null;
				default:
					if (subTest != null)
						throw new SAXException("Missing operator, already got expression");
					subTest = getSubTest(endToken);
					break;
			}
			if (debugTests)
				System.err.println("getTest, subTest = " + subTest.getClass().getName());
		}
		if (debugTests)
			System.err.println("getTest.leave(" + endToken + ", " + isFirstTest + ")");
		return subTest;
	}

	protected ExpressionBase getSubTest(int endToken) throws SAXException
	{
		if (token.type == endToken)
			throw new SAXException("Expected non-empty subexpression");
		switch (token.type) {
			case PathCompilerToken.END:
				throw new SAXException("At end of xpath inside test");
			case PathCompilerToken.ATTRIBUTE:
				return new AttributeExpression(token.txt);
			case PathCompilerToken.PARAN_START:
				return getTest(PathCompilerToken.PARAN_END, false);
			case PathCompilerToken.NOT:
				nextToken(true);
				return new NotExpression(getSubTest(endToken));
			case PathCompilerToken.STRING:
				return new ConstantStringExpression(token.txt);
			default:
				throw new SAXException("Misplaced xpath component type (" + token.type + ")");
		}
	}


	protected void nextToken(boolean insideTest) throws SAXException
	{
		skipWhitespace();
		int startIndex = nextCharIndex;
		doNextToken(insideTest);
		if (debugTokens) {
			String s = "nextToken (" + startIndex + " -> " + nextCharIndex
					+ "): ";
			switch (token.type) {
				case PathCompilerToken.STRING:
					s += "string = '" + token.txt + "'";
					break;
				case PathCompilerToken.INTEGER:
					s += "integer = " + token.integerValue;
					break;
				default:
					s += token.getName();
					if (token.txt != null)
						s += ", " + token.txt;
			}
			System.err.println(s);
		}
	}

	protected void doNextToken(boolean insideTest) throws SAXException
	{
		token.reset();
		if (atEnd()) {
			token.type = PathCompilerToken.END;
			return;
		}
		char ch = xpath.charAt(nextCharIndex++);
		if (debugTokens && false)
			System.err.println("   first token char = " + ch);
		switch (ch) {
			case'[':
				token.type = PathCompilerToken.TEST_START;
				return;
			case']':
				token.type = PathCompilerToken.TEST_END;
				return;
			case'*':
				token.type = PathCompilerToken.ALL_CHILDS;
				return;
			case'=':
				token.type = PathCompilerToken.EQUALS;
				return;
			case'(':
				token.type = PathCompilerToken.PARAN_START;
				return;
			case')':
				token.type = PathCompilerToken.PARAN_END;
				return;
			case'\'':
				parseStringConstant();
				return;
			case'!':
				twoCharOp('=', PathCompilerToken.UNKNOWN, PathCompilerToken.NOT_EQUALS);
				return;
			case'/':
				twoCharOp('/', PathCompilerToken.PATH_SEPARATOR, PathCompilerToken.DEEP_CHILDS);
				return;
			case'.':
				twoCharOp('.', PathCompilerToken.CURRENT_NODE, PathCompilerToken.PARENT_ELEMENT);
				return;
			case'@':
				parseAttributeName();
				return;
		}
		if (ch >= '0' && ch <= '9') {
			parseIntegerConstant();
			return;
		}
		int firstCharIndex = nextCharIndex - 1;
		parseXmlName();
		int len = nextCharIndex - firstCharIndex;
		if (len <= 0)
			throw new SAXException("Can't parse expression");
		String txt = xpath.substring(firstCharIndex, nextCharIndex);
		if (insideTest) {
			if (txt.equals("and")) {
				token.type = PathCompilerToken.AND;
				return;
			}
			if (txt.equals("not")) {
				token.type = PathCompilerToken.NOT;
				return;
			}
			throw new SAXException("Unknown operand inside test");
		}
		token.type = PathCompilerToken.NAMED_CHILD;
		token.txt = xpath.substring(firstCharIndex, nextCharIndex);
	}

	protected void twoCharOp(char ch2, int op1, int op2)
	{
		if (!atEnd() && xpath.charAt(nextCharIndex) == ch2) {
			token.type = op2;
			nextCharIndex++;
		} else {
			token.type = op1;
		}
		return;
	}

	protected void parseStringConstant() throws SAXException
	{
		StringBuilder sb = new StringBuilder();
		int startIndex = nextCharIndex;
		while (true) {
			if (nextCharIndex >= xpathLen)
				throw new SAXException("Quoted string not terminated from index " + startIndex + ": " +
						xpath.substring(startIndex - 1));
			char ch = xpath.charAt(nextCharIndex++);
			if (ch == '\'')
				break;
			if (ch != '\\') {
				sb.append(ch);
				continue;
			}
			ch = xpath.charAt(nextCharIndex++);
			switch (ch) {
				case'n':
					sb.append('\n');
					break;
				case'r':
					sb.append('\r');
					break;
				case't':
					sb.append('\t');
					break;
				default:
					sb.append(ch);
					break;
			}
		}
		token.type = PathCompilerToken.STRING;
		token.txt = sb.toString();
	}

	protected void parseIntegerConstant()
	{
		nextCharIndex--; // need first digit again.
		int value = 0;
		while (true) {
			char ch = xpath.charAt(nextCharIndex++);
			if (ch < '0' || ch > '9')
				break;
			value = value * 10 + (ch - '0');
		}
		token.type = PathCompilerToken.INTEGER;
		token.integerValue = value;
	}

	protected void parseAttributeName()
	{
		int firstCharIndex = nextCharIndex; // omit '@'
		parseXmlName();
		token.type = PathCompilerToken.ATTRIBUTE;
		token.txt = xpath.substring(firstCharIndex, nextCharIndex);
	}

	protected void parseXmlName()
	{
		for (; nextCharIndex < xpathLen; nextCharIndex++) {
			char ch = xpath.charAt(nextCharIndex);
			if ((ch >= 'A' && ch <= 'Z')
					|| (ch >= 'a' && ch <= 'z')
					|| (ch >= '0' && ch <= '9')
					|| ch == '-' || ch == '_' || ch > 128) {
				continue;
			}
			if (ch == ':' && nextCharIndex + 1 < xpathLen
					&& xpath.charAt(nextCharIndex + 1) != ':') {
				continue; // name-space colon.
			}
			break;
		}
	}

	protected final boolean atEnd()
	{
		return nextCharIndex >= xpathLen;
	}

	protected void skipWhitespace()
	{
		while (!atEnd()) {
			char ch = xpath.charAt(nextCharIndex);
			if (ch != ' ' && ch != '\t' && ch != '\n')
				return;
			nextCharIndex++;
		}
	}
}

// END OF FILE
