/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "FilterRuleEdit.h"
#import "FilterRuleEditUtils.h"
#import "GBDebug.h"
#import "XmlDoc.h"
#import "Xml.h"
#import "ServerCommunication.h"
#import "AlertHelper.h"
#import "Prefs.h"
#import "FilterTabController.h"
#import "Rule.h"
#import "Filter.h"
#import "JavascriptSyntaxHighlighter.h"
#import "ClickActionTextField.h"
#import "NamedEnum.h"
#import "AutoZap.h"


#ifdef DEBUG
//#define DEBUG_TEXT_CHANGE
//#define DEBUG_RESPONDS_TO_SELECTOR
//#define DEBUG_DCBS
#endif

static NSFont* regexpCommentFont = NULL;
static NSFont* regexpErrorFont = NULL;

@interface FilterRuleEdit ()

@property FilterTabController* filterTabController;
@property Prefs* prefs;
@property AlertHelper* alertHelper;
@property ServerCommunication* comm;
@property CodeEditHelper* codeEditHelper;
@property NSBundle* bundle;
@property NSXMLElement* ruleE;
@property NSArray* filterArray;
@property Filter* ownerFilter;
@property BOOL virgin;
@property BOOL editable;
@property NSData* previewedData;
@property FilterRuleFieldEditor* fieldEditor;
@property NSString* cachedHostPathQueryValidationPathArgs;
@property XmlDoc* cachedHostPathQueryValidationXmlDoc;
@property NSArray *topLevelObjects;



@property (weak) IBOutlet NSWindow*				ruleSheet;

//----------------------------------------------------------------- header
@property (weak) IBOutlet NSTextField*			filterLabel;
@property (weak) IBOutlet NSPopUpButton*		filterPopupBtn;
@property (weak) IBOutlet NSButton*				enableCB;
@property (weak) IBOutlet NSTabView*			tabView;
@property (weak) IBOutlet NSTextField*			globalErrorTextView;

//----------------------------------------------------------------- setup
@property (strong) IBOutlet NSTabViewItem*		setupTabViewItem;
@property (weak) IBOutlet NSButton*				mobileSafariBtn;
@property (weak) IBOutlet NSPopUpButton*		priorityPopup;
@property (weak) IBOutlet NSMatrix*				actionMatrix;

@property (weak) IBOutlet NSView*				hostPathQueryContainer;
@property (weak) IBOutlet NSTextField*			hostLabel;
@property (weak) IBOutlet NSPopUpButton*		hostTypePopup;
@property (weak) IBOutlet NSTextField*			hostErrorTextField;
@property (weak) IBOutlet FilterRuleTextField*	hostTextField;
@property (weak) IBOutlet NSImageView*			hostErrorLocationImageView;

@property (weak) IBOutlet NSTextField*			pathLabel;
@property (weak) IBOutlet NSPopUpButton*		pathTypePopup;
@property (weak) IBOutlet NSTextField*			pathErrorTextField;
@property (weak) IBOutlet FilterRuleTextField*	pathTextField;
@property (weak) IBOutlet NSImageView*			pathErrorLocationImageView;

@property (weak) IBOutlet NSTextField*			queryLabel;
@property (weak) IBOutlet NSPopUpButton*		queryTypePopup;
@property (weak) IBOutlet NSTextField*			queryErrorTextField;
@property (weak) IBOutlet FilterRuleTextField*	queryTextField;
@property (weak) IBOutlet NSImageView*			queryErrorLocationImageView;

@property (weak) IBOutlet NSView*				keywordContainer;
@property (weak) IBOutlet NSPopUpButton*		keywordTypePopup;
@property (weak) IBOutlet FilterRuleTextField*	keywordTextField;
@property (weak) IBOutlet NSImageView*			keywordErrorLocationImageView;
@property (weak) IBOutlet NSTextField*			keywordErrorTextView;
@property (weak) IBOutlet NSMatrix*				keywordTypeMatrix;
@property (weak) IBOutlet FilterRuleTextField*	keywordUrlField;
@property (weak) IBOutlet ClickActionTextField*	keywordHelpTextLink;

@property (weak) IBOutlet NSView*				commentsContainer;
@property IBOutlet NSTextView*					commentsTextView;

@property (weak) IBOutlet NSTabViewItem*		dummyTabViewItem;

//----------------------------------------------------------------- request
@property (strong) IBOutlet NSTabViewItem*		requestTabViewItem;
@property IBOutlet NSTextView*					requestTextView;
@property (weak) IBOutlet NSButton*				requestHelpBtn;
@property (weak) IBOutlet ClickActionTextField*	requestHelpTextLink;

//----------------------------------------------------------------- css
@property (strong) IBOutlet NSTabViewItem*		cssTabViewItem;
@property IBOutlet NSTextView*					cssTextView;

//----------------------------------------------------------------- javascript
@property (strong) IBOutlet NSTabViewItem*		jsTabViewItem;
@property IBOutlet NSTextView*					jsTextView;
@property (weak) IBOutlet NSPopUpButton*		jsPlacementPop;

//----------------------------------------------------------------- transform
@property (strong) IBOutlet NSTabViewItem*		transformTabViewItem;
@property IBOutlet NSTextView*					transformTextView;
@property (weak) IBOutlet NSPopUpButton*		transformContentTypePop;
@property (weak) IBOutlet NSButton*				transformHelpBtn;
@property (weak) IBOutlet ClickActionTextField*	transformHelpTextLink;

//----------------------------------------------------------------- keyword
@property (strong) IBOutlet NSTabViewItem*		expansionTabViewItem;
@property IBOutlet NSTextView*					expansionTextView;
@property (weak) IBOutlet ClickActionTextField*	expansionHelpTextLink;

//----------------------------------------------------------------- footer
@property (weak) IBOutlet NSButton*				saveBtn;
@property (weak) IBOutlet NSButton*				cancelBtn;
@property (weak) IBOutlet NSButton*				previewBtn;


- (void)makeFilterPopMenu;
- (void)disableControls:(NSView*)view;

- (BOOL)isJavascriptTextStorage:(NSTextStorage*)storage;

- (BOOL)validate:(id)sender;
- (void)saveSettingsInRuleElement:(NSXMLElement*)ruleE;
@property (readonly, copy) NSData *previewDataForCurrentSettings;

@end

@implementation FilterRuleEdit

const static int ACTION_BLOCK = 0;
const static int ACTION_WHITELIST = 1;
const static int ACTION_REQUEST = 2;
const static int ACTION_KEYWORD = 3;

static NamedEnum* hostTypeHelper;
static NamedEnum* pathTypeHelper;
static NamedEnum* queryTypeHelper;
static NamedEnum* keywordTypeHelper;
static NamedEnum* jsPlacementHelper;
static NamedEnum* transformContentTypeHelper;

- (instancetype)initWithFilterTabController:(FilterTabController*)filterTabController
							withOwnerFilter:(Filter*)ownerFilter
							withRuleElement:(NSXMLElement*)ruleE
						 withSuspectElement:(NSXMLElement*)suspectE
{
	if (!(self = [super init]))
		return nil;
	[AutoZap addSticky:self];
	if (!regexpCommentFont) {
		regexpCommentFont = [NSFont systemFontOfSize:9];
		regexpErrorFont = [NSFont boldSystemFontOfSize:9];
		hostTypeHelper = [[NamedEnum alloc] initWithSemicolonList:@"+ignored;is;domain;regexp"];
		pathTypeHelper = [[NamedEnum alloc] initWithSemicolonList:@"+ignored;starts-with;ends-with;contains;not-contains;is;not-is;regexp;not-regexp"];
		queryTypeHelper = [[NamedEnum alloc] initWithSemicolonList:@"+ignored;contains;not-contains;regexp;not-regexp"];
		keywordTypeHelper = [[NamedEnum alloc] initWithSemicolonList:@"+is;starts-with;regexp"];
		jsPlacementHelper = [[NamedEnum alloc] initWithSemicolonList:@"head-start;head-end;body-start;+body-end"];
		transformContentTypeHelper = [[NamedEnum alloc] initWithSemicolonList:@"+html;xml;js;json;css;*;*not-ok*"];
	}
	_topLevelObjects = [AutoZap loadNibNamed:@"FilterRuleEdit" bundle:[filterTabController mainBundle] owner:self];
	NSAssert(_ruleSheet, @"_ruleSheet is null");
	[_ruleSheet setDelegate:self];
	_filterTabController = filterTabController;
	_comm = [filterTabController comm];
	_prefs = [_comm prefs];
	_alertHelper = [_prefs alertHelper];
	_bundle = [filterTabController bundle];
	_ownerFilter = ownerFilter;
	_codeEditHelper = [[CodeEditHelper alloc] initWithDelegate:self];
	//
	if (ruleE) {
		NSAssert(_ownerFilter, @"_ownerFilter");
		if ([_ownerFilter safetySimpleRulesOnly] && ![Rule ruleIsSafe:ruleE])
			[_enableCB setTitle:@"Rule enabled (but ignored)"];
		_virgin = NO;
		_ruleE = ruleE;
		[_saveBtn setTitle:@"Save"];
	} else {
		_ruleE = [Xml createElementWithName:@"rule"];
		_virgin = YES;
	}
	[_enableCB setState:![Xml getBoolAttribute:ruleE withName:@"disabled"]];
	[_mobileSafariBtn setState:[Xml getBoolAttribute:ruleE withName:@"mobile"]];
	int p = [Xml getIntAttribute:ruleE withName:@"priority"];
	p = (p > 0 ? p : 2) - 1;
	[_priorityPopup selectItemAtIndex:MIN(p, [_priorityPopup numberOfItems] - 1)];
	NSString* type = [Xml getAttribute:ruleE withName:@"type"];
	if ([type isEqual:@"keyword"]) {
		[_actionMatrix selectCellAtRow:ACTION_KEYWORD column:0];
	} else {
		[_actionMatrix removeRow:ACTION_KEYWORD]; // keywords not supported in current safari
		if ([type isEqual:@"request"])
			[_actionMatrix selectCellAtRow:ACTION_REQUEST column:0];
		else if (!type || [type isEqual:@"block"])
			[_actionMatrix selectCellAtRow:ACTION_BLOCK column:0];
		else
			[_actionMatrix selectCellAtRow:ACTION_WHITELIST column:0];
	}
	[_commentsTextView setString:soft([Xml stringForXPath:@"comments" withRoot:_ruleE])];
	[_commentsTextView setAlignment:NSLeftTextAlignment range:NSMakeRange(0, [[_commentsTextView string] length])];
	[_keywordTextField setStringValue:soft([Xml getAttribute:_ruleE withName:@"keyword"])];
	NSArray* keywordsE = [_ruleE elementsForName:@"keyword"];
	if ([keywordsE count]) {
		NSXMLElement* keywordE = keywordsE[0];
		[_keywordUrlField setStringValue:soft([Xml getAttribute:keywordE withName:@"keyword-url"])];
		[keywordTypeHelper updatePopButton:_keywordTypePopup fromXPath:@"@keyword-type" withRoot:_ruleE];
		[_keywordTypeMatrix selectCellAtRow:([Xml getBoolAttribute:keywordE withName:@"keyword-uses-js"] ? 0 : 1) column:0];
	} else {
		[_keywordUrlField setStringValue:@""];
		[keywordTypeHelper selectDefaultItemInPop:_keywordTypePopup];
		[_keywordTypeMatrix selectCellAtRow:1 column:0];
	}
	//-------------
	NSString* host = [Xml getAttribute:_ruleE withName:@"host"];
	if (host && [host hasPrefix:@"*."]) {
		[_hostTextField setStringValue:[host substringFromIndex:2]];
		[hostTypeHelper updatePopButton:_hostTypePopup toValue:@"domain"];
	} else {
		[_hostTextField setStringValue:soft(host)];
		if (_virgin || [host length])
			[hostTypeHelper updatePopButton:_hostTypePopup fromXPath:@"@host-type" withRoot:_ruleE withDefault:@"is"];
		else
			[hostTypeHelper selectDefaultItemInPop:_hostTypePopup];
	}
	//-------------
	NSString* path = [Xml getAttribute:_ruleE withName:@"path"];
	[_pathTextField setStringValue:soft(path)];
	if ([path length])
		[pathTypeHelper updatePopButton:_pathTypePopup fromXPath:@"@path-type" withRoot:_ruleE withDefault:@"starts-with"];
	else
		[pathTypeHelper selectDefaultItemInPop:_pathTypePopup];
	//-------------
	NSString* query = [Xml getAttribute:_ruleE withName:@"query"];
	[_queryTextField setStringValue:soft(query)];
	if ([query length])
		[queryTypeHelper updatePopButton:_queryTypePopup fromXPath:@"@query-type" withRoot:_ruleE withDefault:@"contains"];
	else
		[queryTypeHelper selectDefaultItemInPop:_queryTypePopup];
	//-------------
	[_codeEditHelper setCodeTextView:_requestTextView withText:[Xml stringForXPath:@"request" withRoot:_ruleE]];
	[_codeEditHelper setCodeTextView:_cssTextView withText:[Xml stringForXPath:@"css" withRoot:_ruleE]];
	[_codeEditHelper setCodeTextView:_jsTextView withText:[Xml stringForXPath:@"js" withRoot:_ruleE]];
	[_codeEditHelper setCodeTextView:_transformTextView withText:[Xml stringForXPath:@"transform" withRoot:_ruleE]];
	[_codeEditHelper setCodeTextView:_expansionTextView withText:[Xml stringForXPath:@"keyword" withRoot:_ruleE]];
	[self makeFilterPopMenu];
	[jsPlacementHelper updatePopButton:_jsPlacementPop
							   fromXPath:@"js/@placement"
								withRoot:_ruleE];
	[transformContentTypeHelper updatePopButton:_transformContentTypePop
										fromXPath:@"transform/@content-type"
										 withRoot:_ruleE];
	if (suspectE) {
		[_hostTextField setStringValue:[Xml getAttribute:suspectE withName:@"host"]];
		[_pathTextField setStringValue:[Xml getAttribute:suspectE withName:@"path"]];
		[_queryTextField setStringValue:[Xml getAttribute:suspectE withName:@"query"]];
	}
	_editable = [_prefs hasGlimmerAuth] && ![_ownerFilter isSubscription];
	DebugNSLog(@"_editable = %d", _editable);
	if (!_editable) {
		NSView* view = _saveBtn;
		while ([view superview])
			view = [view superview];
		[self disableControls:view];
	}
	NSInteger tabIndex = 0;
	if (!_virgin) {
		tabIndex = [_tabView indexOfTabViewItemWithIdentifier:[_prefs stringForXPath:@"iu/rule/@selected-tab"]];
		if (tabIndex == NSNotFound)
			tabIndex = 0;
	}
	[_tabView setDelegate:self];
	[_tabView selectTabViewItemAtIndex:tabIndex];
	//
	[_keywordTextField setRuleController:self];
	[_keywordTextField setDelegate:self];
	[_keywordUrlField setRuleController:self];
	[_keywordUrlField setDelegate:self];
	//
	[_hostTextField setRuleController:self];
	[_hostTextField setDelegate:self];
	[_pathTextField setRuleController:self];
	[_pathTextField setDelegate:self];
	[_queryTextField setRuleController:self];
	[_queryTextField setDelegate:self];
	[_commentsTextView setDelegate:self];
	//
	[_requestTextView setDelegate:self];
	[[_requestTextView layoutManager] setAllowsNonContiguousLayout:YES];
	[[_requestTextView textStorage] setDelegate:self];
	[_codeEditHelper syntaxHighlight:[_requestTextView textStorage] isJavascript:YES];
	//
	[_cssTextView setDelegate:self];
	[[_cssTextView layoutManager] setAllowsNonContiguousLayout:YES];
	[[_cssTextView textStorage] setDelegate:self];
	//
	[_jsTextView setDelegate:self];
	[[_jsTextView layoutManager] setAllowsNonContiguousLayout:YES];
	[[_jsTextView textStorage] setDelegate:self];
	[_codeEditHelper syntaxHighlight:[_jsTextView textStorage] isJavascript:YES];
	//
	[_transformTextView setDelegate:self];
	[[_transformTextView layoutManager] setAllowsNonContiguousLayout:YES];
	[[_transformTextView textStorage] setDelegate:self];
	[_codeEditHelper syntaxHighlight:[_transformTextView textStorage] isJavascript:YES];
	//
	[_expansionTextView setDelegate:self];
	[[_expansionTextView layoutManager] setAllowsNonContiguousLayout:YES];
	[[_expansionTextView textStorage] setDelegate:self];
	[_codeEditHelper syntaxHighlight:[_expansionTextView textStorage] isJavascript:YES];
	//
	[_requestHelpTextLink setTarget:self selector:@selector(requestHelpAction:)];
	[_transformHelpTextLink setTarget:self selector:@selector(transformHelpAction:)];
	[_expansionHelpTextLink setTarget:self selector:@selector(expansionHelpAction:)];
	[_keywordHelpTextLink setTarget:self selector:@selector(expansionHelpAction:)];
	//
	//
	[_prefs zapRulePreviewElement];
	_previewedData = [self previewDataForCurrentSettings];
	//
	[[_hostPathQueryContainer superview] addSubview:_keywordContainer];
	[_tabView removeTabViewItem:_dummyTabViewItem];
	[_keywordContainer setFrame:[_hostPathQueryContainer frame]];
	//
	if (![_ruleSheet setFrameUsingName:@"FilterEditSheet"]) {
		NSRect r = [_ruleSheet frame];
		r.size.height = 547;
		[_ruleSheet setFrame:r display:YES];
	}
	[self updateDependentStates:self];
	 // until sheet is dismissed.
	[NSApp beginSheet: _ruleSheet
	   modalForWindow: [_alertHelper window]
		modalDelegate: nil
	   didEndSelector: nil
		  contextInfo: nil];
	if ([_tabView selectedTabViewItem] == _setupTabViewItem) {
		[self performSelectorOnMainThread:@selector(updateDependentStates:)
							   withObject:_actionMatrix
							waitUntilDone:NO];
	}
	return self;
}

- (void)dealloc
{
	//DebugNSLog(@"Dealloc RuleEdit");
	[_hostTextField setRuleController:NULL];
	[_pathTextField setRuleController:NULL];
	[_queryTextField setRuleController:NULL];
	// dummyTabViewItem is already released.
}

- (void)makeFilterPopMenu
{
	[_filterPopupBtn removeAllItems];
	/*
	 System font is not provided in italic (only bold) [WTF], so this doesn't work anyway.
	 NSFont* font = [NSFont menuFontOfSize:[NSFont systemFontSize]];
	 font = [[NSFontManager sharedFontManager] convertFont:font toHaveTrait:NSBoldFontMask];
	 NSMutableDictionary* styleDict = [NSMutableDictionary dictionaryWithObject:font forKey:NSFontAttributeName];
	 NSMutableAttributedString* attributedTitle =
		[[NSMutableAttributedString alloc] initWithString:@"New bbbbbb" attributes:styleDict];
	 [[blockerSetPopupBtn itemAtIndex:0] setAttributedTitle:attributedTitle];
	 */
	NSInteger selectedFilterId = _ownerFilter ? [_ownerFilter filterId] : [_prefs intForXPath:@"iu/rule/@default-filter"];
	NSMutableArray* a = [NSMutableArray arrayWithCapacity:10];
	BOOL didSelectFilter = NO;
	NSUInteger lastPrivateFilterIdx = NSNotFound;
	for (Filter* filter in [_filterTabController filterArray]) {
		[a addObject:filter];
		[_filterPopupBtn addItemWithTitle:[filter name]];
		NSUInteger idx = [_filterPopupBtn numberOfItems] - 1;
		if ([filter isSubscription])
			[[_filterPopupBtn itemAtIndex:idx] setEnabled:NO];
		else
			lastPrivateFilterIdx = idx;
		if (selectedFilterId && [filter filterId] == selectedFilterId) {
			[_filterPopupBtn selectItemAtIndex:idx];
			didSelectFilter = YES;
		}
	}
	_filterArray = a;
	[_filterPopupBtn addItemWithTitle:@"+ Create new filter"];
	if (didSelectFilter)
		return;
	if (lastPrivateFilterIdx != NSNotFound)
		[_filterPopupBtn selectItemAtIndex:lastPrivateFilterIdx];
	else
		[_filterPopupBtn selectItemAtIndex:[_filterPopupBtn numberOfItems] - 1];
}

- (void)disableControls:(NSView*)view level:(int)level
{
	//NSLog(@"disable.%d: %@  -- %@", level,
	//	  NSStringFromClass([view class]), [view isKindOfClass:[NSButton class]] ? [((NSButton*)view) title] : @"");
	if (view == _cancelBtn)
		return;
	if ([view isKindOfClass:[NSButton class]]) {
		NSButton* btn = (NSButton*)view;
		if ([btn bezelStyle] == NSHelpButtonBezelStyle)
			return; // don't disable help buttons
		[(NSButton*)view setEnabled:NO];
		return;
	}
	if ([view isKindOfClass:[ClickActionTextField class]])
		return; // don't disable help links
	if ([view isKindOfClass:[NSControl class]]) {
		[(NSControl*)view setEnabled:NO];
		return;
	}
	if ([view isKindOfClass:[NSText class]]) {
		[(NSText*)view setEditable:NO];
		return;
	}
	if ([view isKindOfClass:[NSTabView class]]) {
		// NSTabViewItem is not a subview, so special iterating is required.
		for (NSTabViewItem* item in [(NSTabView*)view tabViewItems])
			[self disableControls:[item view]];
		return;
	}
	for (NSView* sub in [view subviews])
		[self disableControls:sub level:level + 1];
}

- (void)disableControls:(NSView*)view
{
	[self disableControls:view level:0];
}

#pragma mark ------------------------------------- validation

- (void)showErrorImage:(NSImageView*)imageView
		   atCharIndex:(NSUInteger)charIndex
		  errorMessage:(NSString*)errorMessage
			 forString:(NSAttributedString*)as
		  forTextField:(FilterRuleTextField*)textField
		 errorTextView:(NSTextField*)errorTextView
{
	[errorTextView setHidden:NO];
	[errorTextView setTextColor:[NSColor redColor]];
	[errorTextView setFont:regexpErrorFont];
	[errorTextView setStringValue:errorMessage];
	//
	CGFloat w1 = 0, w2 = 0; // get middle of problematic char.
	NSUInteger len = [as length];
	if (charIndex < len)
		as = [as attributedSubstringFromRange:NSMakeRange(0, charIndex + 1)];
	w2 = [as size].width;
	if (charIndex <= len)
		as = [as attributedSubstringFromRange:NSMakeRange(0, charIndex)];
	w1 = [as size].width;
	NSRect frame = [imageView frame];
	CGFloat x = [textField frame].origin.x + 4;
	x -= frame.size.width / 2;
	x += (w1 + w2) / 2;
	x = (CGFloat)round(x);
	[imageView setFrameOrigin:NSMakePoint(x, frame.origin.y)];
	[imageView setHidden:NO];
}

- (BOOL)hasSelectedRegexpInPopMenu:(NSPopUpButton*)typePopup
						 popHelper:(NamedEnum*)popHelper
{
	NSInteger idx = [typePopup indexOfSelectedItem];
	if (idx < 0)
		return NO;
	return [[popHelper nameFromIndex:idx] rangeOfString:@"regexp"].location != NSNotFound;
}

- (void)showErrorImage:(NSImageView*)imageView
		   atCharIndex:(NSString*)charIndex
		  errorMessage:(NSString*)errorMessage
			 forString:(NSAttributedString*)as
		  forTextField:(FilterRuleTextField*)textField
			 typePopup:(NSPopUpButton*)typePopup
			 popHelper:(NamedEnum*)popHelper
		 errorTextView:(NSTextField*)errorTextView
		 regexpComment:(NSString*)regexpComment
{
	BOOL isRegExp = [self hasSelectedRegexpInPopMenu:typePopup popHelper:popHelper];
	[textField setIsRegExp:isRegExp];
	if (![charIndex length]) {
		[imageView setHidden:YES];
		[errorTextView setHidden:!isRegExp];
		if (isRegExp) {
			[errorTextView setTextColor:[NSColor grayColor]];
			[errorTextView setFont:regexpCommentFont];
			[errorTextView setStringValue:regexpComment];
		}
		return;
	}
	[self showErrorImage:imageView
			 atCharIndex:MAX(0, [charIndex intValue])
			errorMessage:errorMessage
			   forString:as
			forTextField:textField
		   errorTextView:errorTextView];
}

- (NSAttributedString*)addValidateItem:(NSPopUpButton*)popup
							 popHelper:(NamedEnum*)popHelper
							 textField:(FilterRuleTextField*)textField
								  name:(NSString*)name
							  pathArgs:(NSMutableString*)pathArgs
{
	NSAttributedString* as = [textField attributedStringValue];
	NSString* type = [popHelper selectedPopValue:popup];
	if (!type) {
		[textField setEnabled:NO];
		return as;
	}
	[textField setEnabled:_editable];
	if (![as length])
		return as;
	BOOL hasQ = [pathArgs rangeOfString:@"?"].location != NSNotFound;
	[pathArgs appendString:(hasQ ? @"&" : @"?")];
	[pathArgs appendString:name];
	[pathArgs appendString:@"-type="];
	[pathArgs appendString:type];
	[pathArgs appendString:@"&"];
	[pathArgs appendString:name];
	[pathArgs appendString:@"="];
	[pathArgs appendString:[GBDebug percentEscapeUrlParameterValue:[as string]]];
	return as;
}

- (BOOL)validateHostPathQuery
{
	//NSDate* st0 = [NSDate date];
	NSMutableString* pathArgs = [NSMutableString stringWithString:@"/filters/validate-edit"];
	NSAttributedString* host = [self addValidateItem:_hostTypePopup
										   popHelper:hostTypeHelper
										   textField:_hostTextField
												name:@"host"
											pathArgs:pathArgs];
	NSAttributedString* path = [self addValidateItem:_pathTypePopup
										   popHelper:pathTypeHelper
										   textField:_pathTextField
												name:@"path"
											pathArgs:pathArgs];
	NSAttributedString* query = [self addValidateItem:_queryTypePopup
											popHelper:queryTypeHelper
											textField:_queryTextField
												 name:@"query"
											 pathArgs:pathArgs];
	NSString* actionDesc = @"Validating host/path/query";
	NSError* __autoreleasing error = NULL;
	XmlDoc* doc;
	if ([_cachedHostPathQueryValidationPathArgs isEqual:pathArgs]) {
		doc = _cachedHostPathQueryValidationXmlDoc;
	} else {
		doc = [_comm curlToXml:pathArgs withActionDescription:actionDesc withErrorRef:&error];
		if (!doc) {
			NSLog(@"Can't validate regexp: %@", error);
			return NO;
		}
		_cachedHostPathQueryValidationPathArgs = pathArgs;
		_cachedHostPathQueryValidationXmlDoc = doc;
	}
	[self showErrorImage:_hostErrorLocationImageView
			 atCharIndex:[doc stringForXPath:@"@host-index"]
			errorMessage:[doc stringForXPath:@"@host-error-message"]
			   forString:host
			forTextField:_hostTextField
			   typePopup:_hostTypePopup
			   popHelper:hostTypeHelper
		   errorTextView:_hostErrorTextField
		   regexpComment:@"must match the complete hostname"];
	[self showErrorImage:_pathErrorLocationImageView
			 atCharIndex:[doc stringForXPath:@"@path-index"]
			errorMessage:[doc stringForXPath:@"@path-error-message"]
			   forString:path
			forTextField:_pathTextField
			   typePopup:_pathTypePopup
			   popHelper:pathTypeHelper
		   errorTextView:_pathErrorTextField
		   regexpComment:@"must match the complete path"];
	[self showErrorImage:_queryErrorLocationImageView
			 atCharIndex:[doc stringForXPath:@"@query-index"]
			errorMessage:[doc stringForXPath:@"@query-error-message"]
			   forString:query
			forTextField:_queryTextField
			   typePopup:_queryTypePopup
			   popHelper:queryTypeHelper
		   errorTextView:_queryErrorTextField
		   regexpComment:@"must match the complete query string"];
	return ![doc boolForXPath:@"@has-error"];
}

- (void)setTextFieldFocus:(FilterRuleTextField*)textField
				 forPopup:(NSPopUpButton*)popup
				forSender:(id)sender
{
	if (sender != popup || [_ownerFilter isSubscription])
		return;
	BOOL used = [popup indexOfSelectedItem] > 0;
	BOOL active = !![textField currentEditor];
	if (used && !active) {
		// doesn't work unless post-poned.
		[_ruleSheet performSelectorOnMainThread:@selector(makeFirstResponder:)
									withObject:textField
								 waitUntilDone:NO];
		return;
	}
	if (!used && active) {
		[_ruleSheet performSelectorOnMainThread:@selector(makeFirstResponder:)
									withObject:_commentsTextView
								 waitUntilDone:NO];
	}
}

- (IBAction)updateDependentStates:(id)sender // needed as validate returns BOOL and not void.
{
	[self validate:sender];
}

- (void)getPreviewData:(NSData**)data ruleElementRef:(NSXMLElement**)ruleRef
{
	NSXMLElement* ruleE = [Xml createElementWithName:@"rule"];
	if (ruleRef)
		*ruleRef = ruleE;
	[self saveSettingsInRuleElement:ruleE];
	if (!_virgin)
		[Xml setAttribute:[Xml getAttribute:_ruleE withName:@"rule-id"] withName:@"rule-id" inElement:ruleE];
	*data = [Xml subtreeToData:ruleE];
}

- (NSData*)previewDataForCurrentSettings
{
	NSData* data;
	[self getPreviewData:&data ruleElementRef:NULL];
	return data;
}

- (void)setTextViewEditable:(NSTextView*)textView
{
	[textView setEditable:_editable];
	[textView setSelectable:YES];
	if (textView == _cssTextView) {
		NSColor* color = _editable ? [NSColor blackColor] : [NSColor grayColor];
		[textView setTextColor:color];
	}
}

- (void)makeTabItem:(NSTabViewItem*)tabItem visible:(BOOL)wantedVisible atIndex:(int*)tabIndex
{
	NSAssert(tabItem, @"tabItem is NULL");
	BOOL isVisible = ([_tabView indexOfTabViewItem:tabItem] != NSNotFound);
	//DebugNSLog(@"makeTabItem (%d/%d = %@) visible %d -> %d", *tabIndex, [tabView numberOfTabViewItems], [tabItem label], isVisible, wantedVisible);
	if (wantedVisible) {
		if (!isVisible)
			[_tabView insertTabViewItem:tabItem atIndex:*tabIndex];
		(*tabIndex)++;
	} else if (isVisible) {
		[_tabView removeTabViewItem:tabItem];
	}
}

-(void)setTabItem:(NSTabViewItem*)tabItem withLabelDot:(BOOL)wantsDot
{
	NSString* s = [tabItem label];
	NSString* dot = @"•\u2009"; // 2009 is thin space
	BOOL hasDot = [s hasPrefix:dot];
	if (hasDot && !wantsDot)
		[tabItem setLabel:[s substringFromIndex:[dot length]]];
	else if (!hasDot && wantsDot)
		[tabItem setLabel:[dot stringByAppendingString:s]];
}

- (BOOL)validateKeywordText
{
	NSString* s = [_keywordTextField stringValue];
	NSInteger len = [s length];
	if (!len) {
		[_keywordErrorLocationImageView setHidden:YES];
		[_keywordErrorTextView setStringValue:@"Keyword must be specified."];
		return NO;
	}
	NSInteger type = [_keywordTypePopup indexOfSelectedItem];
	if (type == 2) {
		NSMutableString* pathArgs = [NSMutableString stringWithString:@"/filters/validate-edit"];
		NSAttributedString* keyword = [self addValidateItem:_keywordTypePopup
												  popHelper:keywordTypeHelper
												  textField:_keywordTextField
													   name:@"keyword"
												   pathArgs:pathArgs];
		NSString* actionDesc = @"Validating keyword";
		NSError*__autoreleasing error = NULL;
		XmlDoc* doc = [_comm curlToXml:pathArgs withActionDescription:actionDesc withErrorRef:&error];
		if (!doc) {
			NSLog(@"Can't validate regexp: %@", error);
			return NO;
		}
		[self showErrorImage:_keywordErrorLocationImageView
				 atCharIndex:[doc stringForXPath:@"@keyword-index"]
				errorMessage:[doc stringForXPath:@"@keyword-error-message"]
				   forString:keyword
				forTextField:_keywordTextField
				   typePopup:_keywordTypePopup
				   popHelper:keywordTypeHelper
			   errorTextView:_keywordErrorTextView
			   regexpComment:@"must match the complete keyword"];
		return ![doc boolForXPath:@"@has-error"];
	}
	for (int i = 0; i < len; i++) {
		unichar ch = [s characterAtIndex:i];
		if ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || (ch >= '0' && ch <= '9') || ch == '-')
			continue;
		NSString* msg;
		if (ch == ' ')
			msg = @"Keywords must not contain spaces.";
		else
			msg = @"Invalid character. Only a-z, A-Z, 0-9 and dash (-) are allowed.";
		[self showErrorImage:_keywordErrorLocationImageView
				 atCharIndex:i
				errorMessage:msg
				   forString:[_keywordTextField attributedStringValue]
				forTextField:_keywordTextField
			   errorTextView:_keywordErrorTextView];
		return NO;
	}
	[_keywordErrorLocationImageView setHidden:YES];
	BOOL useJS = ![_keywordTypeMatrix selectedRow];
	if (type == 1 && !useJS) {
		[_keywordErrorTextView setStringValue:@"Starts-with is useful only for Javascript based expansion."];
		return NO;
	}
	[_keywordErrorTextView setStringValue:@""];
	return YES;
}

- (BOOL)validateKeyword:(id)sender  tabIndex:(int*)tabIndex
{
	[_hostErrorLocationImageView setHidden:YES];
	[_pathErrorLocationImageView setHidden:YES];
	[_queryErrorLocationImageView setHidden:YES];
	if (sender == _actionMatrix) {
		[_ruleSheet performSelectorOnMainThread:@selector(makeFirstResponder:)
									withObject:_keywordTextField
								 waitUntilDone:NO];
	}
	BOOL useJS = ![_keywordTypeMatrix selectedRow];
	[_keywordUrlField setHidden:useJS];
	[self makeTabItem:_expansionTabViewItem visible:useJS atIndex:tabIndex];
	BOOL valid = [self validateKeywordText];
	NSString* urlError = NULL;
	if (useJS) {
		BOOL hasExpansion = !![_codeEditHelper trimmedTextViewString:_expansionTextView];
		[self setTabItem:_expansionTabViewItem withLabelDot:hasExpansion];
		if (!hasExpansion)
			urlError = @"Javascript missing in 'expansion' tab.";
	} else {
		NSString* u = [_keywordUrlField stringValue];
		if (![u length]) {
			urlError = @"URL must be specified.";
		} else {
			NSString* u2 = [u stringByReplacingOccurrencesOfString:@"%s" withString:@"dummy"];
			u2 = [u2 stringByReplacingOccurrencesOfString:@"%S" withString:@"dummy"];
			u2 = [u2 stringByReplacingOccurrencesOfString:@"%t" withString:@"dummy"];
			u2 = [u2 stringByReplacingOccurrencesOfString:@"|" withString:@"z"]; // NSUrl doesn't accept '|' in URLs.
			@try {
				NSURL* url = [NSURL URLWithString:u2];
				NSString* scheme = [url scheme];
				if (![scheme length])
					urlError = @"URL is missing scheme";
				else if (([scheme isEqual:@"http"] || [scheme isEqual:@"https"]) && ![[url host] length])
					urlError = @"URL is missing hostname";
				else if ([u2 isEqual:u])
					urlError = @"URL is missing %s or %S";
			}
			@catch (NSException *ex) {
				urlError = @"URL is invalid.";
			}
		}
	}
	[_globalErrorTextView setStringValue:soft(urlError)];
	valid &= !urlError;
	[_saveBtn setEnabled:_editable && valid];
	[_previewBtn setEnabled:_editable && valid && ![_previewedData isEqualToData:[self previewDataForCurrentSettings]]];
	//--------------------------------------------------------------------------------
	return valid;
}

- (BOOL)validate:(id)sender
{
	NSInteger action = [_actionMatrix selectedRow];
	BOOL isModification = (action == ACTION_WHITELIST);
	BOOL isRequest = (action == ACTION_REQUEST);
	BOOL isKeyword = (action == ACTION_KEYWORD);
	int tabIndex = 1;
	[self makeTabItem:_requestTabViewItem visible:isRequest atIndex:&tabIndex];
	[self makeTabItem:_cssTabViewItem visible:isModification atIndex:&tabIndex];
	[self makeTabItem:_jsTabViewItem visible:isModification atIndex:&tabIndex];
	[self makeTabItem:_transformTabViewItem visible:isModification atIndex:&tabIndex];
	//
	[_keywordContainer setHidden:!isKeyword];
	[_hostPathQueryContainer setHidden:isKeyword];
	//
	[_hostLabel setHidden:isKeyword];
	[_hostTypePopup setHidden:isKeyword];
	[_hostErrorTextField setHidden:isKeyword];
	[_hostTextField setHidden:isKeyword];
	//
	[_pathLabel setHidden:isKeyword];
	[_pathTypePopup setHidden:isKeyword];
	[_pathErrorTextField setHidden:isKeyword];
	[_pathTextField setHidden:isKeyword];
	//
	[_queryLabel setHidden:isKeyword];
	[_queryTypePopup setHidden:isKeyword];
	[_queryErrorTextField setHidden:isKeyword];
	[_queryTextField setHidden:isKeyword];
	//
	[self setTextViewEditable:_requestTextView];
	[self setTextViewEditable:_cssTextView];
	[self setTextViewEditable:_jsTextView];
	[self setTextViewEditable:_transformTextView];
	[self setTextViewEditable:_expansionTextView];
	//
	if (isKeyword)
		return [self validateKeyword:sender tabIndex:&tabIndex];
	[self makeTabItem:_expansionTabViewItem visible:NO atIndex:&tabIndex];
	//
	[_keywordErrorLocationImageView setHidden:YES];
	[self setTextFieldFocus:_hostTextField forPopup:_hostTypePopup forSender:sender];
	[self setTextFieldFocus:_pathTextField forPopup:_pathTypePopup forSender:sender];
	[self setTextFieldFocus:_queryTextField forPopup:_queryTypePopup forSender:sender];
	if (sender == _actionMatrix) {
		NSView* dst;
		if ([_hostTextField isEnabled])
			dst = _hostTextField;
		else if ([_pathTextField isEnabled])
			dst = _pathTextField;
		else if ([_queryTextField isEnabled])
			dst = _queryTextField;
		else
			dst = _commentsTextView;
		[_ruleSheet performSelectorOnMainThread:@selector(makeFirstResponder:)
									withObject:dst
								 waitUntilDone:NO];
	}
	//--------------------------------------------------------------------------------
	BOOL hostPathQueryAreValid = [self validateHostPathQuery];
	BOOL valid = _editable && hostPathQueryAreValid;
	//--------------------------------------------------------------------------------
	BOOL hasRequestScript = !![_codeEditHelper trimmedTextViewString:_requestTextView];
	BOOL css = isModification && !![_codeEditHelper trimmedTextViewString:_cssTextView];
	BOOL js = isModification && !![_codeEditHelper trimmedTextViewString:_jsTextView];
	BOOL transform = isModification && !![_codeEditHelper trimmedTextViewString:_transformTextView];
	[self setTabItem:_requestTabViewItem withLabelDot:isRequest && hasRequestScript];
	[self setTabItem:_cssTabViewItem withLabelDot:!!css];
	[self setTabItem:_jsTabViewItem withLabelDot:!!js];
	[self setTabItem:_transformTabViewItem withLabelDot:!!transform];
	//--------------------------------------------------------------------------------
	[_globalErrorTextView setStringValue:(isRequest && !hasRequestScript) ? @"Missing script in 'request' tab" : @""];
	valid &= !isRequest || hasRequestScript;
	//--------------------------------------------------------------------------------
	[_saveBtn setEnabled:valid];
	[_previewBtn setEnabled:valid && ![_previewedData isEqualToData:[self previewDataForCurrentSettings]]];
	//--------------------------------------------------------------------------------
	return valid;
}

- (void)textDidChange:(NSNotification *)aNotification
{
#ifdef DEBUG_TEXT_CHANGE
	DebugNSLog(@"textDidChange: %@", aNotification);
#endif
	[self updateDependentStates:self];
}

- (void)controlTextDidChange:(NSNotification *)aNotification
{
#ifdef DEBUG_TEXT_CHANGE
	DebugNSLog(@"controlTextDidChange: %@", aNotification);
#endif
	[self updateDependentStates:self];
}

- (void)tabView:(NSTabView *)tabView didSelectTabViewItem:(NSTabViewItem *)tabViewItem
{
#ifdef DEBUG_TEXT_CHANGE
	DebugNSLog(@"didSelectTabViewItem: %@", tabViewItem);
#endif
	[self updateDependentStates:self];
	[_codeEditHelper delayedSyntaxHighlightJavascript:[_requestTextView textStorage]];
	[_codeEditHelper delayedSyntaxHighlightJavascript:[_jsTextView textStorage]];
	[_codeEditHelper delayedSyntaxHighlightJavascript:[_transformTextView textStorage]];
}

#pragma mark --------------------------- URL paste

- (BOOL)handlePasteUrl:(NSString*)txt withTextField:(NSTextField*)textField
{
	BOOL isHostField = (textField == _hostTextField);
	BOOL isPathField = (textField == _pathTextField);
	if (!isHostField && !isPathField)
		return NO;
	if ([Xml getHttpUrlSyntaxError:txt])
		return NO;
	NSURL* url;
	@try {
		url = [NSURL URLWithString:txt];
	}
	@catch (NSException* ex) {
		return NO;
	}
	NSString* host = [url host];
	if (![host length])
		return NO;
	if (isHostField)
		[_hostTextField setStringValue:host];
	NSString* path = [url path];
	BOOL needValidate = NO;
	if ([path length] && ![path isEqual:@"/"] && (isPathField || ![[_pathTextField stringValue] length])) {
		if (![_pathTypePopup indexOfSelectedItem])
			[_pathTypePopup selectItemAtIndex:1];
		[_pathTextField setEnabled:YES];
		[_pathTextField setStringValue:path];
		needValidate = YES;
		if (isHostField) {
			[_ruleSheet performSelectorOnMainThread:@selector(makeFirstResponder:)
										withObject:_pathTextField
									 waitUntilDone:NO];
		}
	} else {
		[self updateDependentStates:self];
	}
	NSString* query = [url query];
	if ([query length] && ![[_queryTextField stringValue] length])
		[_queryTextField setStringValue:query];
	if (needValidate)
		[self updateDependentStates:self];
	return YES;
}

#pragma mark --------------------------- prettyprænt

- (void)syntaxHighlightRegExp:(NSMutableAttributedString*)attributedString
{
	[_codeEditHelper syntaxHighlight:attributedString isJavascript:NO];
}

- (BOOL)isJavascriptTextStorage:(NSTextStorage*)storage
{
	return storage != [_cssTextView textStorage];
}

- (void)textStorageDidProcessEditing:(NSNotification *)noti
{
	[_codeEditHelper textStorageDidProcessEditing:noti];
}

- (IBAction)requestHelpAction:(id)sender
{
	NSURL* url = [NSURL URLWithString:@"http://glimmerblocker.org/wiki/ModifyRequest"];
	[[NSWorkspace sharedWorkspace] openURL:url];
}

- (IBAction)transformHelpAction:(id)sender
{
	NSURL* url = [NSURL URLWithString:@"http://glimmerblocker.org/wiki/Transform2"];
	[[NSWorkspace sharedWorkspace] openURL:url];
}

- (IBAction)expansionHelpAction:(id)sender
{
	NSURL* url = [NSURL URLWithString:@"http://glimmerblocker.org/wiki/KeywordExpansion"];
	[[NSWorkspace sharedWorkspace] openURL:url];
}

#pragma mark --------------------------- committing etc.

- (void)saveSelectedTabInPrefs
{
	NSTabViewItem* tab = [_tabView selectedTabViewItem];
	if (!tab)
		return;
	[_prefs setString:[tab identifier] forXPath:@"iu/rule/@selected-tab"];
	[_prefs markAsDirty];
}

- (void)dismissSheet
{
	[_ruleSheet saveFrameUsingName:@"FilterEditSheet"];
	[NSApp endSheet:_ruleSheet];
	[_ruleSheet orderOut:self];
	[AutoZap zapSticky:self];
}

- (IBAction)cancelAction:(id)sender
{
	if (!_virgin)
		[self saveSelectedTabInPrefs];
	[self dismissSheet];
	if ([_prefs zapRulePreviewElement])
		[_comm saveDirtyPrefsAndNotify];
	[AutoZap add:self];
}

- (NSXMLElement*)storeCDataText:(NSString*)txt withName:(NSString*)name inElement:(NSXMLElement*)parentE
{
	if (![txt length])
		return NULL;
	NSXMLElement* e = [Xml createElementWithName:name];
	[parentE addChild:e];
	NSXMLNode* cdata = [[NSXMLNode alloc] initWithKind:NSXMLTextKind options:NSXMLNodeIsCDATA];
	[cdata setStringValue:txt];
	[e addChild:cdata];
	return e;
}

- (void)saveTextField:(NSTextField*)textField
			 attrName:(NSString*)attrName
			 withType:(NSString*)type
		 typeAttrName:(NSString*)typeAttrName
			inElement:(NSXMLElement*)e
{
	NSCharacterSet* ws = [NSCharacterSet whitespaceAndNewlineCharacterSet];
	NSString* value = [[textField stringValue] stringByTrimmingCharactersInSet:ws];
	if (!type || ![value length])
		type = value = NULL;
	[Xml setAttribute:value withName:attrName inElement:e];
	[Xml setAttribute:type withName:typeAttrName inElement:e];
}

- (void)saveHostPathQueryInRuleElement:(NSXMLElement*)ruleE
{
	[self saveTextField:_hostTextField
			   attrName:@"host"
			   withType:[hostTypeHelper selectedPopValue:_hostTypePopup]
		   typeAttrName:@"host-type"
			  inElement:ruleE];
	[self saveTextField:_pathTextField
			   attrName:@"path"
			   withType:[pathTypeHelper selectedPopValue:_pathTypePopup]
		   typeAttrName:@"path-type"
			  inElement:ruleE];
	[self saveTextField:_queryTextField
			   attrName:@"query"
			   withType:[queryTypeHelper selectedPopValue:_queryTypePopup]
		   typeAttrName:@"query-type"
			  inElement:ruleE];
}

- (void)saveSettingsInRuleElement:(NSXMLElement*)ruleE
{
	[Xml zapChildsAndAttributes:ruleE];
	//
	[Xml setBool:![_enableCB state] withName:@"disabled" inElement:ruleE];
	[Xml setBool:NO withName:@"enabled" inElement:ruleE]; // old style.
	[Xml setBool:[_mobileSafariBtn state] withName:@"mobile" inElement:ruleE];
	[Xml setAttribute:[NSString stringWithFormat:@"%ld", [_priorityPopup indexOfSelectedItem] + 1] withName:@"priority" inElement:ruleE];
	NSInteger action = [_actionMatrix selectedRow];
	BOOL isWhitelist = (action == ACTION_WHITELIST);
	BOOL isRequest = (action == ACTION_REQUEST);
	BOOL isKeyword = (action == ACTION_KEYWORD);
	//--------------------------------------------------------------------------------
	[self storeCDataText:[_codeEditHelper trimmedTextViewString:_commentsTextView]
				withName:@"comments"
			   inElement:ruleE];
	if (isKeyword) {
		[Xml setAttribute:@"keyword" withName:@"type" inElement:ruleE];
		[self saveTextField:_keywordTextField
				   attrName:@"keyword"
				   withType:[keywordTypeHelper selectedPopValue:_keywordTypePopup]
			   typeAttrName:@"keyword-type"
				  inElement:ruleE];
		NSXMLElement* keywordE = [self storeCDataText:[_codeEditHelper trimmedTextViewString:_expansionTextView]
											 withName:@"keyword"
											inElement:ruleE];
		if (keywordE) {
			[Xml setAttribute:@"js" withName:@"language" inElement:keywordE];
			[Xml setAttribute:@"1" withName:@"version" inElement:keywordE];
		} else {
			keywordE = [Xml createElementWithName:@"keyword"];
			[ruleE addChild:keywordE];
		}
		[Xml setBool:([_keywordTypeMatrix selectedRow] ? NO : YES) withName:@"keyword-uses-js" inElement:keywordE];
		[Xml setAttribute:[_keywordUrlField stringValue] withName:@"keyword-url" inElement:keywordE];
	} else if (isRequest) {
		[Xml setAttribute:@"request" withName:@"type" inElement:ruleE];
		[self saveHostPathQueryInRuleElement:ruleE];
		NSXMLElement* requestE = [self storeCDataText:[_codeEditHelper trimmedTextViewString:_requestTextView]
											 withName:@"request"
											inElement:ruleE];
		if (requestE) {
			[Xml setAttribute:@"js" withName:@"language" inElement:requestE];
			[Xml setAttribute:@"1" withName:@"version" inElement:requestE];
		}
	} else {
		[self saveHostPathQueryInRuleElement:ruleE];
		//--------------------------------------------------------------------------------
		//
		NSXMLElement* cssE = [self storeCDataText:[_codeEditHelper trimmedTextViewString:_cssTextView]
										 withName:@"css"
										inElement:ruleE];
		//
		NSXMLElement* jsE = [self storeCDataText:[_codeEditHelper trimmedTextViewString:_jsTextView]
										withName:@"js"
									   inElement:ruleE];
		if (jsE) {
			[Xml setAttribute:[jsPlacementHelper selectedPopValue:_jsPlacementPop]
					 withName:@"placement"
					inElement:jsE];
		}
		//
		NSXMLElement* transformE = [self storeCDataText:[_codeEditHelper trimmedTextViewString:_transformTextView]
											   withName:@"transform"
											  inElement:ruleE];
		if (transformE) {
			[Xml setAttribute:@"js" withName:@"language" inElement:transformE];
			[Xml setAttribute:@"1" withName:@"version" inElement:transformE];
			NSString* ct = [transformContentTypeHelper selectedPopValue:_transformContentTypePop];
			[Xml setAttribute:ct withName:@"content-type" inElement:transformE];
		}
		//
		NSString* type;
		if (!isWhitelist)
			type = @"block";
		else if (cssE || jsE || transformE)
			type = @"modify";
		else
			type = @"whitelist";
		[Xml setAttribute:type withName:@"type" inElement:ruleE];
		// @whitelist is obsolete attribute, but keep for backwards compability with 1.1
		[Xml setBool:isWhitelist withName:@"whitelist" inElement:ruleE];
	}
	//--------------------------------------------------------------------------------
	[Xml setAttribute:NULL withName:@"comments" inElement:ruleE]; // obsolete.
	[Xml setAttribute:NULL withName:@"path-starts-with" inElement:ruleE]; // obsolete.
	[Xml setAttribute:NULL withName:@"path-ends-with" inElement:ruleE]; // obsolete.
	[Xml setAttribute:NULL withName:@"path-regexp" inElement:ruleE]; // obsolete.
}

- (IBAction)previewAction:(id)sender
{
	if (![self validate:sender])
		return;
	NSXMLElement* rulePreviewE = [_prefs rulePreviewE];
	[Xml zapChildsAndAttributes:rulePreviewE];
	NSXMLElement* ruleE;
	NSData* data;
	[self getPreviewData:&data ruleElementRef:&ruleE];
	_previewedData = data;
	[rulePreviewE addChild:ruleE];
	if (_ownerFilter)
		[Xml setInt:[_ownerFilter filterId] withName:@"for-filter-id" inElement:rulePreviewE];
	[_prefs markAsDirty];
	[_comm saveDirtyPrefsAndNotify];
	[self validate:self];
}

- (void)commitRuleToFilter:(NSInteger)idx wantsNewFilter:(BOOL)wantsNewFilter
{
	Filter* newOwnerFilter;
	if (wantsNewFilter)
		newOwnerFilter = [_filterTabController createNewUntitledFilter];
	else
		newOwnerFilter = _filterArray[idx];
	//DebugNSLog(@"newOwnerFilter = %@", [newOwnerFilter name]);
	if (newOwnerFilter != _ownerFilter) {
		[_ruleE detach];
		[_ownerFilter notifyDirty];
		[Xml setAttribute:NULL withName:@"rule-id" inElement:_ruleE]; // current rule-id might not be unique in new filter.
		_ownerFilter = NULL;
		_ownerFilter = newOwnerFilter;
		[_prefs setInt:[_ownerFilter filterId] forXPath:@"iu/rule/@default-filter"];
	}
	if (![_ruleE parent])
		[[_ownerFilter filterDataE] addChild:_ruleE];
	[FilterRuleEdit getRuleId:_ruleE]; // ensure it has a rule-id
}

- (IBAction)commitAction:(id)sender
{
	NSAssert(!(_virgin && [_ruleE parent]), @"virgin with parent");
	if (![self validate:sender])
		return;
	//
	NSInteger idx = [_filterPopupBtn indexOfSelectedItem];
	NSAssert(idx >= 0, @"Missing filter in pop");
	BOOL wantsNewFilter = idx >= (int)[_filterArray count];
	BOOL filterUpdated = wantsNewFilter || idx != (int)[_filterArray indexOfObject:_ownerFilter];
	//
	[_prefs zapRulePreviewElement];
	NSData* oldContents = _virgin || filterUpdated ? NULL : [Xml subtreeToData:_ruleE];
	[self saveSettingsInRuleElement:_ruleE];
	BOOL changed = ![oldContents isEqualToData:[Xml subtreeToData:_ruleE]];
	DebugNSLog(changed ? @"Rule changed." : @"Rule NOT changed.");
	if (changed)
		[self commitRuleToFilter:idx wantsNewFilter:wantsNewFilter];
	[self saveSelectedTabInPrefs];
	[self dismissSheet];
	if (changed)
		[_filterTabController ruleEditCommit:_ruleE withFilter:_ownerFilter isVirgin:_virgin];
	else
		[_prefs saveIfDirty:self];
	[AutoZap add:self];
}

- (void)selectTextView:(NSView*)view
{
	if ([view isKindOfClass:[NSTextField class]])
		[view setFocusRingType:NSFocusRingTypeExterior]; // NSTextView/NSScrollView can't display focus ring.
	[_ruleSheet makeFirstResponder:view];
}

- (BOOL)handleTab:(NSView *)control delta:(int)delta
{
	NSMutableArray* a = [NSMutableArray array];
	if ([_actionMatrix selectedRow] == ACTION_KEYWORD) {
		[a addObject:_keywordTextField];
		if ([_keywordTypeMatrix selectedRow])
			[a addObject:_keywordUrlField];
		DebugNSLog(@"handleTab.kw: %@", a);
	} else {
		if ([_hostTextField isEnabled])
			[a addObject:_hostTextField];
		if ([_pathTextField isEnabled])
			[a addObject:_pathTextField];
		if ([_queryTextField isEnabled])
			[a addObject:_queryTextField];
	}
	[a addObject:_commentsTextView];
	if ([a count] <= 1)
		return YES;
	NSInteger current = [a indexOfObject:control];
	NSInteger idx;
	if (current == NSNotFound) {
		DebugNSLog(@"Did not find current tab-view in available views.");
		idx = 0;
	} else {
		idx = (current + [a count] + delta) % [a count];
	}
	[control setFocusRingType:NSFocusRingTypeNone];
	if ([control isKindOfClass:[NSTextField class]]) {
		NSText* t = [(NSTextField*)control currentEditor];
		[t setSelectedRange:NSMakeRange(0, 0)];
	}
	[self performSelectorOnMainThread:@selector(selectTextView:)
						   withObject:a[idx]
						waitUntilDone:NO];
	return YES;
}

- (BOOL)textView:(NSTextView*)aTextView doCommandBySelector:(SEL)aSelector
{
#ifdef DEBUG_DCBS
	DebugNSLog(@"doCommandBySelector: %s", aSelector);
#endif
	if (aTextView == _commentsTextView) {
		if (aSelector == @selector(insertNewline:)) {
			[_commentsTextView insertNewlineIgnoringFieldEditor:self];
			return YES; // Yes, I did handle it.
		}
		if (aSelector == @selector(insertTab:))
			return [self handleTab:_commentsTextView delta:1];
		if (aSelector == @selector(insertBacktab:))
			return [self handleTab:_commentsTextView delta:-1];
	}
	BOOL newlineSelector = (aSelector == @selector(insertNewline:));
	// I can't find noop: anywhere, so I don't know what the expected return type is.
	BOOL noopSelector = [NSStringFromSelector(aSelector) isEqual:@"noop:"];
	if (!newlineSelector && !noopSelector)
		return NO; // No, I did not handle it.
	NSEvent* event = [NSApp currentEvent];
	if ([event type] != NSKeyDown || [event isARepeat])
		return NO; // No, I did not handle it.
#ifdef DEBUG_DCBS
	DebugNSLog(@" current keydown event: %@", [NSApp currentEvent]);
#endif
	NSUInteger modifiers = [event modifierFlags] & NSDeviceIndependentModifierFlagsMask;
	NSString* s = [event charactersIgnoringModifiers];
	BOOL enter = newlineSelector && !modifiers && [s length] == 1 && [s characterAtIndex:0] == 3;
	BOOL appleReturn = noopSelector && (modifiers == NSCommandKeyMask) && [s length] == 1 && [s characterAtIndex:0] == 13;
#ifdef DEBUG_DCBS
	DebugNSLog(@"KeyDown(%d = %x) for %d, enter = %d, appleReturn = %d", modifiers, modifiers, [s characterAtIndex:0], enter, appleReturn);
#endif
	if (!enter && !appleReturn)
		return NO; // No, I did not handle it.
#ifdef DEBUG_DCBS
	DebugNSLog(@"KeyDown executes self.commitAction:");
#endif
	[_saveBtn highlight:YES];
	[self performSelectorOnMainThread:@selector(commitAction:)
						   withObject:self
						waitUntilDone:NO];
	return YES; // Yes, I did handle it.
}

- (BOOL)control:(NSControl *)control textView:(NSTextView *)textView doCommandBySelector:(SEL)command
{
#ifdef DEBUG_DCBS
	DebugNSLog(@"control.textView.dCBS: %@  for %@", NSStringFromSelector(command), (control == hostTextField ? @"host" : @"?"));
#endif
	// dunno why tab doesn't work. Hack so it does.
	if (command == @selector(insertTab:))
		return [self handleTab:control delta:1];
	if (command == @selector(insertBacktab:))
		return [self handleTab:control delta:-1];
	if (command == @selector(insertNewline:) && [_saveBtn isEnabled]) {
		// dunno why this doesn't work the first time return is hit while being in the host field.
		[_saveBtn highlight:YES];
		[self performSelectorOnMainThread:@selector(commitAction:)
							   withObject:self
							waitUntilDone:NO];
		return YES;
	}
	return NO;
}

- (id)windowWillReturnFieldEditor:(NSWindow *)window toObject:(id)anObject
{
	if (![anObject isKindOfClass:[FilterRuleTextField class]])
		return NULL;
	//DebugNSLog(@"windowWillReturnFieldEditor:  %s", object_getClassName(anObject));
	if (!_fieldEditor)
		_fieldEditor = [[FilterRuleFieldEditor alloc] initWithFrame:NSMakeRect(0, 0, 1, 1)];
	return _fieldEditor;
}

+ (int)getRuleId:(NSXMLElement*)ruleE
{
	int ruleId = [Xml getIntAttribute:ruleE withName:@"rule-id"];
	if (!ruleId) {
		ruleId = [Xml makeUniqueIdWithRoot:(NSXMLElement*)[ruleE parent] withAttributeName:@"rule-id"];
		[Xml setInt:ruleId withName:@"rule-id" inElement:ruleE];
	}
	return ruleId;
}

#ifdef DEBUG_RESPONDS_TO_SELECTOR
- (BOOL)respondsToSelector:(SEL)aSelector
{
	BOOL rts = [super respondsToSelector:aSelector];
	DebugNSLog(@"FilterRuleEdit, %@ respondsToSelector  %s", (rts ? @"does   " : @"does not"), aSelector);
	return rts;
}
#endif

@end

