/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "ProxyScriptEditor.h"
#import "GBDebug.h"
#import "ClickActionTextField.h"
#import "ServerCommunication.h"
#import "Xml.h"
#import "Prefs.h"
#import "NamedEnum.h"
#import "AlertHelper.h"
#import "AutoZap.h"

NamedEnum* proxyTypeEnum;

@interface ProxyScriptEditor ()

@property NSBundle* bundle;
@property id<ProxyScriptEditorDelegate> delegate;
@property Prefs* prefs;
@property AlertHelper* alertHelper;
@property ServerCommunication* comm;
@property CodeEditHelper* codeEditHelper;
@property NSData* oldSettingsData;
@property NSString* helpMessage;
@property NSTimer* timer;
@property NSArray *topLevelObjects;

@property (weak) IBOutlet NSWindow*				scriptSheet;
@property (weak) IBOutlet NSTabView*			tabView;

@property IBOutlet NSTextView*					scriptCodeView;
@property (weak) IBOutlet NSTextField*			scriptMessageField;

@property (weak) IBOutlet NSButton*				scriptHelpBtn;
@property (weak) IBOutlet ClickActionTextField*	scriptHelpTextLink;

@property IBOutlet NSTextView*					networkCodeView;

@property (weak) IBOutlet NSButton*				saveBtn;


- (BOOL)isJavascriptTextStorage:(NSTextStorage*)storage;

@property (readonly) BOOL validateSyntax;

- (void)startTimerWithImmediateExecute:(BOOL)immediate;
- (void)stopTimer;
- (void)timerCallback:(NSTimer*)theTimer;

- (void)updateNetworkCodeView;

@end

@implementation ProxyScriptEditor

+ (void)initProxyScriptEditor
{
	if (proxyTypeEnum)
		return;
	proxyTypeEnum = [[NamedEnum alloc] initWithSemicolonList:@"+none;javascript;static"];
}

- (instancetype)initWithServerCommunication:(ServerCommunication*)comm
					   withBundle:(NSBundle*)bundle
					 withDelegate:(id<ProxyScriptEditorDelegate>)delegate
{
	if (!(self = [super init]))
		return nil;
	[AutoZap addSticky:self];
	_topLevelObjects = [AutoZap loadNibNamed:@"ProxyScriptEdit" bundle:[comm mainBundle] owner:self];
	NSAssert(_scriptSheet, @"scriptSheet is null");
	//
	_bundle = bundle;
	_delegate = delegate;
	_comm = comm;
	_prefs = [_comm prefs];
	_alertHelper = [_prefs alertHelper];
	_codeEditHelper = [[CodeEditHelper alloc] initWithDelegate:self];
	//
	[_tabView selectFirstTabViewItem:self];
	[_tabView setDelegate:self];
	[_scriptHelpTextLink setTarget:self selector:@selector(scriptHelpAction:)];
	_helpMessage = [_scriptMessageField stringValue];
	[_networkCodeView setEditable:NO];
	//
	[_codeEditHelper setCodeTextView:_scriptCodeView
							withText:[_prefs stringForXPath:@"network-settings/proxy-script"]];
	[_scriptCodeView setDelegate:self];
	[[_scriptCodeView layoutManager] setAllowsNonContiguousLayout:YES];
	[[_scriptCodeView textStorage] setDelegate:self];
	[_codeEditHelper syntaxHighlight:[_scriptCodeView textStorage] isJavascript:YES];
	//
	NSXMLElement* scriptE = (NSXMLElement*)[_prefs nodeForXPath:@"network-settings/proxy-script"];
	_oldSettingsData = scriptE ? [Xml subtreeToData:scriptE] : NULL;
	//
	[_scriptSheet setFrameUsingName:@"ProxyScriptEditSheet"];
	 // until sheet is dismissed.
	[NSApp beginSheet: _scriptSheet
	   modalForWindow: [_alertHelper window]
		modalDelegate: nil
	   didEndSelector: nil
		  contextInfo: nil];
	[self startTimerWithImmediateExecute:YES];
	return self;
}

- (void)dealloc
{
	[self stopTimer];
}

- (BOOL)isJavascriptTextStorage:(NSTextStorage*)storage
{
	return YES;
}

- (void)textStorageDidProcessEditing:(NSNotification *)noti
{
	[_codeEditHelper textStorageDidProcessEditing:noti];
	[self startTimerWithImmediateExecute:NO];
}

//====================================================================================

- (void)updateNetworkCodeView
{
	NSString *path = [NSString stringWithFormat:@"%@%@%@",
					  [_bundle bundlePath],
					  @"/Contents/NetworkConfigurationListener.app",
					  @"/Contents/MacOS/NetworkConfigurationListener"];
	NSLog(@"NHT: %@", path);
	NSMutableString *json = [NSMutableString stringWithCapacity:10];
	[GBDebug executeCommandlineTool:path withArgs:NULL stdoutString:json];
	if ([json isEqual:[_networkCodeView string]])
		return; // no need to spend time reformatting code.
	[_networkCodeView setString:json];
	[_codeEditHelper syntaxHighlight:[_networkCodeView textStorage] isJavascript:YES];
}

//====================================================================================

- (void)startTimerWithImmediateExecute:(BOOL)immediate
{
	[self stopTimer];
	_timer = [NSTimer scheduledTimerWithTimeInterval:(immediate ? 0.1 : 1.2)
											   target:self
											 selector:@selector(timerCallback:)
											 userInfo:NULL
											  repeats:NO];
}

- (void)stopTimer
{
	if (!_timer)
		return;
	[_timer invalidate];
	_timer = NULL;
}

- (void)timerCallback:(NSTimer*)theTimer
{
	[self validateSyntax];
}

//====================================================================================

- (BOOL)textView:(NSTextView*)aTextView doCommandBySelector:(SEL)aSelector
{
#ifdef DEBUG_DCBS
	DebugNSLog(@"doCommandBySelector: %s", aSelector);
#endif
	BOOL newlineSelector = (aSelector == @selector(insertNewline:));
	// I can't find noop: anywhere, so I don't know what the expected return type is.
	BOOL noopSelector = [NSStringFromSelector(aSelector) isEqual:@"noop:"];
	if (!newlineSelector && !noopSelector)
		return NO; // No, I did not handle it.
	NSEvent* event = [NSApp currentEvent];
	if ([event type] != NSKeyDown || [event isARepeat])
		return NO; // No, I did not handle it.
#ifdef DEBUG_DCBS
	DebugNSLog(@" current keydown event: %@", [NSApp currentEvent]);
#endif
	NSUInteger modifiers = [event modifierFlags] & NSDeviceIndependentModifierFlagsMask;
	NSString* s = [event charactersIgnoringModifiers];
	BOOL enter = newlineSelector && !modifiers && [s length] == 1 && [s characterAtIndex:0] == 3;
	BOOL appleReturn = noopSelector && (modifiers == NSCommandKeyMask) && [s length] == 1 && [s characterAtIndex:0] == 13;
#ifdef DEBUG_DCBS
	DebugNSLog(@"KeyDown(%d = %x) for %d, enter = %d, appleReturn = %d", modifiers, modifiers, [s characterAtIndex:0], enter, appleReturn);
#endif
	if (!enter && !appleReturn)
		return NO; // No, I did not handle it.
#ifdef DEBUG_DCBS
	DebugNSLog(@"KeyDown executes self.commitAction:");
#endif
	[_saveBtn highlight:YES];
	[self performSelectorOnMainThread:@selector(commitAction:)
						   withObject:self
						waitUntilDone:NO];
	return YES; // Yes, I did handle it.
}

//====================================================================================

- (IBAction)scriptHelpAction:(id)sender
{
	NSURL* url = [NSURL URLWithString:@"http://glimmerblocker.org/wiki/Proxies"];
	[[NSWorkspace sharedWorkspace] openURL:url];
}

- (void)dismissSheet
{
	[_scriptSheet saveFrameUsingName:@"ProxyScriptEditSheet"];
	[NSApp endSheet:_scriptSheet];
	[_scriptSheet orderOut:self];
	[AutoZap zapSticky:self];
}

- (IBAction)cancelAction:(id)sender
{
	[self stopTimer];
	[self dismissSheet];
	[AutoZap add:self];
}

- (BOOL)validateSyntax
{
	NSString* js = [_codeEditHelper trimmedTextViewString:_scriptCodeView];
	NSString* pathArgs = [NSString stringWithFormat:@"/site-proxy/validate?javascript=%@",
						  [GBDebug percentEscapeUrlParameterValue:js]];
	XmlDoc* doc = [_comm curlToXml:pathArgs withActionDescription:NULL];
	if (!doc) {
		DebugNSLog(@"Problem validating JS.");
		return YES; // don't stop the user saving it, if the proxy failed running.
	}
	NSString* msg = [doc stringForXPath:@"@proxy-error"];
	if ([msg length]) {
		[_scriptMessageField setStringValue:msg];
		[_scriptMessageField setTextColor:[NSColor redColor]];
		return NO;
	} else {
		[_scriptMessageField setStringValue:_helpMessage];
		[_scriptMessageField setTextColor:[NSColor blackColor]];
		return YES;
	}
}

- (void)saveSettingsInElement:(NSXMLElement*)scriptE
{
	NSString* js = [_codeEditHelper trimmedTextViewString:_scriptCodeView];
	//DebugNSLog(@"Script: {{{%@}}}", js);
	if (![js length])
		return;
	[Xml setCData:js forXPath:@"." withRoot:scriptE];
	[Xml setAttribute:@"js" withName:@"language" inElement:scriptE];
	[Xml setAttribute:@"1" withName:@"version" inElement:scriptE];
}

- (BOOL)isDirty
{
	NSXMLElement* scriptE = [Xml createElementWithName:@"proxy-script"];
	[self saveSettingsInElement:scriptE];
	NSData* data = [Xml subtreeToData:scriptE];
	return ![data isEqualToData:_oldSettingsData];
}

- (IBAction)commitAction:(id)sender
{
	[self stopTimer];
	BOOL dirty = [self isDirty];
	if (dirty) {
		NSXMLElement* scriptE = [_prefs singletonElementForXPath:@"network-settings/proxy-script"];
		[self saveSettingsInElement:scriptE];
		[_prefs markAsDirty];
		[_comm saveDirtyPrefsAndNotify];
	}
	[self dismissSheet];
	[_delegate notifyProxyScriptEditorCommitted:self];
	[AutoZap add:self];
}

@end

@implementation ProxyScriptEditor(NSTabViewDelegate)

- (void)tabView:(NSTabView *)tabView didSelectTabViewItem:(NSTabViewItem *)tabViewItem
{
	if ([[tabViewItem identifier] isEqual:@"network"])
		[self updateNetworkCodeView];
}

@end
