/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "VerticalViewFlow.h"
#import "GBDebug.h"

#ifdef DEBUG
//#defined DEBUG_TABLEVIEW 1
#endif

@implementation VerticalViewFlowRow

- (instancetype)initWithFrame:(NSRect)frameRect
{
	if (!(self = [super initWithFrame:frameRect]))
		return nil;
	return self;
}

- (void)setRowNumber:(NSUInteger)rowNumber
{
	if (_rowNumber == rowNumber)
		return;
	_rowNumber = rowNumber;
	[self setNeedsDisplay:YES];
}

- (NSUInteger)rowNumber
{
	return _rowNumber;
}

- (void)setSelected:(BOOL)selected
{
	if (_selected == selected)
		return;
	_selected = selected;
	if (!selected)
		[self endEditing];
	[self setNeedsDisplay:YES];
}

- (BOOL)selected
{
	return _selected;
}

- (void)drawRect:(NSRect)r
{
	if (_selected) {
		if ([self shouldDrawUsingFocusColor])
			[[NSColor alternateSelectedControlColor] set];
		else
			[[NSColor secondarySelectedControlColor] set];
	} else {
		NSArray* a = [NSColor controlAlternatingRowBackgroundColors];
		NSUInteger idx = _rowNumber % [a count];
		NSColor* color = a[idx];
		[color set];
	}
	NSRectFill(r);
}

- (BOOL)shouldDrawUsingFocusColor
{
	if (!_selected)
		return NO;
	NSWindow* window = [self window];
	if (![window isKeyWindow])
		return NO;
	return [self isEditing] || [[self tableView] isFirstResponder];
}

- (BOOL)isEditing
{
	return NO;
}

- (void)endEditing
{
}

- (VerticalViewFlow*)tableView
{
	NSView* view = [self superview];
	while (view) {
		if ([view isKindOfClass:[VerticalViewFlow class]])
			return (VerticalViewFlow*)view;
		view = [view superview];
	}
	return NULL;
}

@end

@interface VerticalViewFlow()

@property BOOL isFirstResponder;
@property CGFloat rowBottomY;
@property CGFloat lastRowHeight;
@property NSWindow* notificationWindow;
@property NSScrollView* scrollView;

- (void)removeNotifications;
- (void)contentResized:(NSNotification*)notification;

@end

@implementation VerticalViewFlow

- (BOOL) isFlipped
{
	return YES;
}

- (void)findScroller
{
	if (_scrollView)
		return;
	NSView* parent = self;;
	while (1) {
		parent = [parent superview];
		if (!parent)
			return;
		if ([parent isKindOfClass:[NSClipView class]])
			continue;
		if ([parent isKindOfClass:[NSScrollView class]])
			break;
		return;
	}
	_scrollView = (NSScrollView*)parent;
	NSNotificationCenter* nc = [NSNotificationCenter defaultCenter];
	[nc addObserver:self
		   selector:@selector(contentResized:)
			   name:NSViewFrameDidChangeNotification
			 object:_scrollView];
}

- (instancetype)initWithFrame:(NSRect)frameRect
{
	if (!(self = [super initWithFrame:frameRect]))
		return nil;
	_rows = [NSMutableArray arrayWithCapacity:10];
	[self performSelectorOnMainThread:@selector(findScroller) withObject:NULL waitUntilDone:NO];
	return self;
}

- (void)dealloc
{
	if (_scrollView) {
		NSNotificationCenter* nc = [NSNotificationCenter defaultCenter];
		[nc addObserver:self
			   selector:@selector(contentResized:)
				   name:NSViewFrameDidChangeNotification
				 object:_scrollView];
	}
	[self removeNotifications];
}

- (void)contentResized:(NSNotification*)notification
{
	[self findScroller];
	[self updateRowViewFrames];
}

- (void)removeNotifications
{
	if (!_notificationWindow)
		return;
	NSNotificationCenter* nc = [NSNotificationCenter defaultCenter];
	[nc removeObserver:self
				  name:NSWindowDidBecomeKeyNotification
				object:_notificationWindow];
	[nc removeObserver:self
				  name:NSWindowDidResignKeyNotification
				object:_notificationWindow];
	_notificationWindow = NULL;
}

- (void)handleWindowBecameResignedKey:(id)sender
{
	[self repaintSelectedRows];
}

- (void)viewDidMoveToWindow
{
	[self removeNotifications];
	_notificationWindow = [self window];
	if (!_notificationWindow)
		return;
	NSNotificationCenter* nc = [NSNotificationCenter defaultCenter];
	[nc addObserver:self
		   selector:@selector(handleWindowBecameResignedKey:)
			   name:NSWindowDidBecomeKeyNotification
			 object:_notificationWindow];
	[nc addObserver:self
		   selector:@selector(handleWindowBecameResignedKey:)
			   name:NSWindowDidResignKeyNotification
			 object:_notificationWindow];
}

- (void)addRow:(VerticalViewFlowRow*)view atIndex:(NSUInteger)idx
{
	if (idx >= [_rows count])
		[_rows addObject:view];
	else
		[_rows insertObject:view atIndex:idx];
	[self addSubview:view];
	if ([_delegate respondsToSelector:@selector(vertPatchTableAddedRow:)])
		[_delegate vertPatchTableAddedRow:self];
	[self updateRowViewFrames];
}

- (void)removeRow:(VerticalViewFlowRow*)view
{
	[_rows removeObject:view];
	if ([view superview] == self)
		[view removeFromSuperview];
	if ([_delegate respondsToSelector:@selector(vertPatchTableRemovedRow:)])
		[_delegate vertPatchTableRemovedRow:self];
	[self updateRowViewFrames];
}

- (id)rowByIndex:(NSUInteger)idx
{
	return _rows[idx];
}

- (NSUInteger)rowCount
{
	return [_rows count];
}

- (NSArray*)rows
{
	return [NSArray arrayWithArray:_rows];
}

- (void)setFreezeUpdates:(BOOL)freeze
{
	if (freeze == _freezeUpdates)
		return;
	_freezeUpdates = freeze;
	if (!freeze)
		[self updateRowViewFrames];
}

- (void)updateRowViewFrames
{
	if (_freezeUpdates)
		return;
	NSUInteger num = [_rows count];
	CGFloat totalHeight = 0;
	CGFloat maxWidth = 0;
	for (NSUInteger i = 0; i < num; i++) {
		VerticalViewFlowRow* view = _rows[i];
		NSRect r = [view frame];
		totalHeight += r.size.height;
		maxWidth = MAX(maxWidth, r.size.width);
	}
	NSRect parentFrame = [[self superview] frame];
	totalHeight = MAX(totalHeight, parentFrame.size.height);
	NSRect sr = [self frame];
	if (sr.size.width != maxWidth || sr.size.height != totalHeight)
		[self setFrameSize:NSMakeSize(maxWidth, totalHeight)];
#ifdef DEBUG_TABLEVIEW
	DebugNSLog(@"updateRowViewFrames, maxW = %f, totalHeight = %f", maxWidth, totalHeight);
#endif
	CGFloat y = 0;
	for (NSUInteger i = 0; i < num; i++) {
		VerticalViewFlowRow* view = _rows[i];
		NSRect vr = [view frame];
#ifdef DEBUG_TABLEVIEW
		DebugNSLog(@"   origin[%d]:  0, %f   height=%f", i, y, vr.size.height);
#endif
		if (vr.origin.y != y) {
			[view setFrameOrigin:NSMakePoint(0, y)];
			[view setNeedsDisplay:YES];
		}
		[view setRowNumber:i];
		_lastRowHeight = vr.size.height;
		y += vr.size.height;
	}
	_rowBottomY = y;
}

- (void)drawDummyRows
{
	NSArray* a = [NSColor controlAlternatingRowBackgroundColors];
	NSRect r = [self frame];
	CGFloat endY = r.size.height;
	r.origin.y = _rowBottomY;
	r.size.height = _lastRowHeight;
	for (NSUInteger row = [_rows count]; r.origin.y < endY; row++) {
		NSUInteger idx = row % [a count];
		NSColor* color = a[idx];
		[color set];
		NSRectFill(r);
		r.origin.y += r.size.height;
	}
}

- (void)repaintSelectedRows
{
#ifdef DEBUG_TABLEVIEW
	NSLog(@"repaintSelectedRows");
#endif
	NSUInteger num = [_rows count];
	for (NSUInteger i = 0; i < num; i++) {
		VerticalViewFlowRow* view = _rows[i];
		if ([view selected])
			[view setNeedsDisplay:YES];
	}
}

- (void)drawRect:(NSRect)theRect
{
	if (_lastRowHeight && _rowBottomY >= theRect.origin.y)
		[self drawDummyRows];
}

- (void)setSelectionByView:(VerticalViewFlowRow*)selectedView
{
	NSUInteger num = [_rows count];
	BOOL changed = NO;
	for (NSUInteger i = 0; i < num; i++) {
		VerticalViewFlowRow* view = _rows[i];
		BOOL wanted = (view == selectedView);
		if ([view selected] == wanted)
			continue;
		[view setSelected:wanted];
		changed = YES;
	}
	if (!changed)
		return;
	if ([_delegate respondsToSelector:@selector(vertPatchTableChangedSelection:)])
		[_delegate vertPatchTableChangedSelection:self];
}

- (void)setSelectionByIndex:(NSUInteger)idx
{
	[self setSelectionByView:[self rowByIndex:idx]];
}

- (id)firstSelectedRow
{
	NSUInteger num = [_rows count];
	for (NSUInteger i = 0; i < num; i++) {
		VerticalViewFlowRow* view = _rows[i];
		if ([view selected])
			return view;
	}
	return NULL;
}

- (id)lastSelectedRow
{
	NSUInteger num = [_rows count];
	if (num == 0)
		return NULL;
	for (NSUInteger i = num - 1; 1; i--) {
		VerticalViewFlowRow* view = _rows[i];
		if (i == 0 || [view selected])
			return view;
	}
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- (void)keyDown:(NSEvent*)theEvent
{
	NSUInteger mask = NSAlphaShiftKeyMask | NSShiftKeyMask | NSControlKeyMask
		| NSAlternateKeyMask | NSCommandKeyMask;
	if ([theEvent modifierFlags] & mask) {
		[super keyDown:theEvent];
		return;
	}
	NSString* s = [theEvent characters];
	for (NSUInteger si = 0; si < [s length]; si++) {
		unichar ch = [s characterAtIndex:si];
		if (ch == NSUpArrowFunctionKey) {
			VerticalViewFlowRow* row = [self firstSelectedRow];
			if (row && [row rowNumber] > 0)
				[self setSelectionByIndex:[row rowNumber] - 1];
			else if (!row && [_rows count])
				[self setSelectionByIndex:[_rows count] - 1];
		}
		if (ch == NSDownArrowFunctionKey) {
			VerticalViewFlowRow* row = [self lastSelectedRow];
			if (row && [row rowNumber] < [_rows count] - 1)
				[self setSelectionByIndex:[row rowNumber] + 1];
			else if (!row && [_rows count])
				[self setSelectionByIndex:0];
		}
	}
	[super keyDown:theEvent];
}

- (void)updateDrawFocusRing:(BOOL)flag
{
	NSView* view = [self superview];
	while (view) {
		if ([view isKindOfClass:[ScrollViewWithFocusRing class]]) {
			ScrollViewWithFocusRing* scroller = (ScrollViewWithFocusRing*)view;
			[scroller setDrawFocusRing:flag];
			return;
		}
		view = [view superview];
	}
}

- (void)mouseDown:(NSEvent *)theEvent
{
#ifdef DEBUG_TABLEVIEW
	DebugNSLog(@"MouseDown in table view: %@", theEvent);
#endif
	NSUInteger num = [_rows count];
	for (NSUInteger i = 0; i < num; i++) {
		VerticalViewFlowRow* view = _rows[i];
		[view endEditing];
	}
}

- (BOOL)becomeFirstResponder
{
	_isFirstResponder = YES;
#ifdef DEBUG_TABLEVIEW
	NSWindow* window = [self window];
	DebugNSLog(@"TableView.becomeFirstResponder: main = %d, isRF = %d",
		  [window isMainWindow], [window firstResponder] == self);
#endif
	[self repaintSelectedRows];
	[self updateDrawFocusRing:YES];
	return YES;
}

- (BOOL)resignFirstResponder
{
#ifdef DEBUG_TABLEVIEW
	DebugNSLog(@"TableView.resignFirstResponder");
#endif
	_isFirstResponder = NO;
	[self repaintSelectedRows];
	[self updateDrawFocusRing:NO];
	return YES;
}

@end

@implementation ScrollViewWithFocusRing

- (instancetype)initWithFrame:(NSRect)frameRect
{
	if (!(self = [super initWithFrame:frameRect]))
		return nil;
	return self;
}


- (void)setDrawFocusRing:(BOOL)flag
{
	if (flag == _drawFocusRing)
		return;
	_drawFocusRing = flag;
	[self setKeyboardFocusRingNeedsDisplayInRect: [self bounds]];
}

- (void)drawRect:(NSRect)r
{
	[super drawRect:r];
	if (!_drawFocusRing)
		return;
	NSSetFocusRingStyle(NSFocusRingOnly);
	NSRectFill([self bounds]);
}

@end
