/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.rule.AbstractRule;
import org.glimmerblocker.proxy.rule.HostMatchType;
import org.glimmerblocker.proxy.rule.KeywordRule;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * For fast matching of hostnames or domain-wildcards to hostnames.
 * The hostnames are stored using a tree by splitting it at the dots,
 * and using the root domain as the tree root.
 */
public class RuleCollection
{
	@NotNull public final Logger logger;

	private final DomainNode root = new DomainNode();
	private final List<AbstractRule> hostRegexpList = new ArrayList<>();
	private final List<AbstractRule> hostIgnoredList = new ArrayList<>();

	private final List<KeywordRule> keywordList = new ArrayList<>();

	public RuleCollection(@NotNull Logger logger)
	{
		this.logger = logger;
	}

	public synchronized void clear()
	{
		root.subNodes = null;
		root.ruleList = null;
	}

	@NotNull
	public List<KeywordRule> getKeywordRules(@NotNull String keyword)
	{
		List<KeywordRule> list = new ArrayList<>();
		for (KeywordRule rule : keywordList) {
			switch (rule.keywordMatchType) {
				case Is:
					if (!keyword.equals(rule.keyword))
						continue;
					break;
				case StartsWith:
					if (!keyword.startsWith(rule.keyword))
						continue;
					break;
				case Regexp:
					if (!rule.keywordRegexp.matcher(keyword).matches())
						continue;
					break;
			}
			list.add(rule);
		}
		return list;
	}

	@NotNull
	public List<AbstractRule> getRules(@NotNull String host,
	                                   @NotNull String path,
	                                   @NotNull String query,
	                                   boolean mobile)
	{
		List<AbstractRule> matches = new ArrayList<>();
		String pathLC = path.toLowerCase();
		String queryLC = query.toLowerCase();
		getRules(host, pathLC, queryLC, mobile, matches);
		addMatches(hostRegexpList, matches, false, host, pathLC, queryLC, mobile);
		addMatches(hostIgnoredList, matches, false, host, pathLC, queryLC, mobile);
		Collections.sort(matches);
		return matches;
	}

	private void getRules(@NotNull String host,
	                      @NotNull String pathLC,
	                      @NotNull String queryLC,
	                      boolean mobile,
	                      @NotNull List<AbstractRule> matches)
	{
		DomainNode n = root;
		// subhost is the hostname seen from 'n'.  "adserver.adtech.de" becomes "adserver.adtech",
		// which becomes "adserver", and finally null.
		String subhost = host;
		while (true) {
			boolean isDomainMatch = (subhost != null);
			addMatches(n.ruleList, matches, isDomainMatch, host, pathLC, queryLC, mobile);
			if (!isDomainMatch || n.subNodes == null)
				break;
			int idx = subhost.lastIndexOf('.');
			String key = (idx < 0) ? subhost : subhost.substring(idx + 1);
			n = n.subNodes.get(key);
			if (n == null)
				return;
			subhost = (idx < 0) ? null : subhost.substring(0, idx);
		}
	}

	/**
	 * Add the filters for this node which matches the path.
	 */
	void addMatches(@Nullable List<AbstractRule> sourceList,
	                @NotNull List<AbstractRule> matches,
	                boolean isDomainMatch,
	                @NotNull String host,
	                @NotNull String pathLC,
	                @NotNull String queryLC,
	                boolean mobile)
	{
		if (sourceList == null)
			return;
		for (AbstractRule rule : sourceList) {
			if (isDomainMatch && rule.hostMatchType != HostMatchType.Domain)
				continue;
			if (!mobile && rule.mobileOnly)
				continue;
			if (rule.hostMatchType == HostMatchType.Regexp) {
				// all other host types have already been checked.
				if (!rule.hostRegexp.matcher(host).matches())
					continue;
			}
			switch (rule.pathMatchType) {
				case Ignored:
					break;
				case StartsWith:
					if (!pathLC.startsWith(rule.pathLC))
						continue;
					break;
				case EndsWith:
					if (!pathLC.endsWith(rule.pathLC))
						continue;
					break;
				case Is:
					if (!pathLC.equals(rule.pathLC))
						continue;
					break;
				case NotIs:
					if (pathLC.equals(rule.pathLC))
						continue;
					break;
				case Contains:
					if (!pathLC.contains(rule.pathLC))
						continue;
					break;
				case NotContains:
					if (pathLC.contains(rule.pathLC))
						continue;
					break;
				case Regexp:
					if (!rule.pathRegexp.matcher(pathLC).matches())
						continue;
					break;
				case NotRegexp:
					if (rule.pathRegexp.matcher(pathLC).matches())
						continue;
					break;
				default:
					System.err.println("Invalid pathMatchType = " + rule.pathMatchType);
			}
			switch (rule.queryMatchType) {
				case Ignored:
					break;
				case Contains:
					if (!queryLC.contains(rule.queryLC))
						continue;
					break;
				case NotContains:
					if (queryLC.contains(rule.queryLC))
						continue;
					break;
				case Regexp:
					if (!rule.queryRegexp.matcher(queryLC).matches())
						continue;
					break;
				case NotRegexp:
					if (rule.queryRegexp.matcher(queryLC).matches())
						continue;
					break;
				default:
					System.err.println("Invalid queryMatchType = " + rule.queryMatchType);
			}
			matches.add(rule);
		}
	}

	/**
	 * @param rule The filter to add.
	 */
	public synchronized void add(@NotNull AbstractRule rule)
	{
		if (rule instanceof KeywordRule) {
			keywordList.add((KeywordRule)rule);
			return;
		}
		switch (rule.hostMatchType) {
			case Regexp:
				hostRegexpList.add(rule);
				return;
			case Ignored:
				hostIgnoredList.add(rule);
				return;
		}
		String s = rule.host;
		DomainNode n = root;
		while (true) {
			if (s.length() == 0) {
				n.add(rule);
				return;
			}
			int idx = s.lastIndexOf('.');
			String key = (idx < 0) ? s : s.substring(idx + 1);
			if (n.subNodes == null)
				n.subNodes = new ConcurrentHashMap<>();
			DomainNode sub = n.subNodes.get(key);
			if (sub == null) {
				sub = new DomainNode();
				n.subNodes.put(key, sub);
			}
			n = sub;
			s = idx < 0 ? "" : s.substring(0, idx);
		}
	}

	/**
	 * A Node in the tree. Contains a map of sub-nodes and the list of filters for this node.
	 */
	protected static class DomainNode
	{
		@Nullable Map<String, DomainNode> subNodes;
		@Nullable List<AbstractRule> ruleList;

		void add(@NotNull AbstractRule filter)
		{
			if (ruleList == null)
				ruleList = Collections.synchronizedList(new ArrayList<>(1));
			ruleList.add(filter);
		}
	}
}
