/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.network.dns;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.StringTokenizer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Immutable.
 */
public class DnsResolverConfig implements Comparable<DnsResolverConfig>
{
	public final boolean isNativeResolver;

	/**
	 * Empty if no servers specified, and the OS resolver should be used.
	 */
	@NotNull
	public final List<String> serverList;

	/**
	 * Whitespace is normalized.
	 */
	@NotNull
	public final String javascriptText;

	/**
	 * Debug description
	 */
	@NotNull
	public final String description;

	/**
	 * Use the Mac OS X native resolver on localhost.
	 */
	@NotNull
	public static final DnsResolverConfig NATIVE_RESOLVER = new DnsResolverConfig(
			true, Collections.<String>emptyList(), "", "native Mac OS X resolver");


	@NotNull
	public static DnsResolverConfig decode(@Nullable String txt)
	{
		if (txt == null)
			return NATIVE_RESOLVER;
		txt = txt.trim();
		if (txt.length() == 0)
			return NATIVE_RESOLVER;
		List<String> list = new ArrayList<>();
		StringTokenizer st = new StringTokenizer(txt, ",; \n\t");
		StringBuilder sb = new StringBuilder(txt.length() + 5);
		while (st.hasMoreTokens()) {
			String s = st.nextToken().trim();
			if (s.length() <= 0)
				continue;
			list.add(s);
			if (sb.length() > 0)
				sb.append(", ");
			sb.append(s);
		}
		if (list.isEmpty())
			return NATIVE_RESOLVER; // if txt is ",,,"
		list = Collections.unmodifiableList(list);
		String s = sb.toString();
		return new DnsResolverConfig(false, list, s, s);
	}

	private DnsResolverConfig(boolean isNativeResolver,
	                          @NotNull List<String> serverList,
	                          @NotNull String javascriptText,
	                          @NotNull String description)
	{
		this.isNativeResolver = isNativeResolver;
		this.serverList = serverList;
		this.javascriptText = javascriptText;
		this.description = description;
	}

	@Override
	public boolean equals(Object o)
	{
		return (o instanceof DnsResolverConfig)
				&& javascriptText.equals(((DnsResolverConfig)o).javascriptText);
	}

	@Override
	public int hashCode()
	{
		return javascriptText.hashCode();
	}

	public int compareTo(@NotNull DnsResolverConfig k)
	{
		return javascriptText.compareTo(k.javascriptText);
	}
}
