/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.network.dns;

import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.jetbrains.annotations.NotNull;
import org.xbill.DNS.ARecord;
import org.xbill.DNS.Cache;
import org.xbill.DNS.DClass;
import org.xbill.DNS.ExtendedResolver;
import org.xbill.DNS.Lookup;
import org.xbill.DNS.Record;
import org.xbill.DNS.Resolver;
import org.xbill.DNS.SimpleResolver;
import org.xbill.DNS.TextParseException;
import org.xbill.DNS.Type;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.ref.FatRef;
import dk.vitality.lib.connectionCache.ConnectionInfo;

public class DnsResolverConnection implements ConnectionInfo<DnsResolverConnection>
{
	@NotNull private final DnsResolverConfig config;
	@NotNull private final Logger logger;
	@NotNull private FatRef<Boolean> debugNetworkConfig;
	@NotNull private final ExtendedResolver resolver;

	public DnsResolverConnection(@NotNull DnsResolverConfig config,
	                             @NotNull Logger logger,
	                             @NotNull FatRef<Boolean> debugNetworkConfig) throws UnknownHostException
	{
		this.config = config;
		this.logger = logger;
		this.debugNetworkConfig = debugNetworkConfig;
		List<Resolver> resolvers = new ArrayList<>(config.serverList.size());
		for (String s : config.serverList) {
			//noinspection UnusedCatchParameter
			try {
				resolvers.add(new SimpleResolver(s));
			}
			catch (UnknownHostException ex) {
				System.err.println("Unknown DNS server name: '" + s + "'." +
						" You SHOULD use dotted decimal for DNS servers instead of host names.");
			}
		}
		this.resolver = new ExtendedResolver(resolvers.toArray(new Resolver[resolvers.size()]));
	}

	public void debugDump(@NotNull String indent, @NotNull PrintWriter pw)
	{
		pw.print(indent);
		pw.println("DnsResolver: " + config.serverList);
	}

	public boolean isClosed() throws Exception
	{
		return false;
	}

	public boolean isReusable()
	{
		return true;
	}

	public void abort()
	{
	}

	@NotNull
	public DnsResolverConnection getConnection()
	{
		return this;
	}

	public InetAddress[] lookupHost(@NotNull String hostname) throws UnknownHostException
	{
		boolean debugFlag = debugNetworkConfig.get();
		if (debugFlag)
			logger.ln("DNS.getAllByName: '" + hostname + "' using " + config.description);
		long st0 = debugFlag ? System.currentTimeMillis() : 0;
		Lookup helper;
		try {
			helper = new Lookup(hostname, Type.A, DClass.IN);
		}
		catch (TextParseException ex) {
			//noinspection ThrowInsideCatchBlockWhichIgnoresCaughtException
			throw new UnknownHostException(ExceptionUtils.getExMsg(ex));
		}
		helper.setCache(dnsCache);
		helper.setResolver(resolver);
		helper.run();
		long dur = debugFlag ? System.currentTimeMillis() - st0 : 0;
		int result = helper.getResult();
		if (result == Lookup.SUCCESSFUL) {
			Record[] records = helper.getAnswers();
			List<InetAddress> list = new ArrayList<>(records.length);
			for (Record r : records) {
				if (r instanceof ARecord) {
					ARecord ar = (ARecord)r;
					list.add(ar.getAddress());
				}
			}
			if (debugFlag) {
				logger.ln("DNS resolved '" + hostname + "' to: " + list + " in " + dur + " msecs" +
						" using '" + config.description + "'");
			}
			return list.toArray(new InetAddress[list.size()]);
		}
		if (debugFlag)
			logger.ln("DNS failed to resolve '" + hostname + "', time = " + dur + " msecs.");
		switch (result) {
			case Lookup.HOST_NOT_FOUND:
				throw new UnknownHostException("Host '" + hostname + "' not found");
			case Lookup.TYPE_NOT_FOUND:
				throw new UnknownHostException("Host '" + hostname + "' not found (no A record)");
			case Lookup.UNRECOVERABLE:
				throw new UnknownHostException("Cannot lookup host '" + hostname + "'");
			case Lookup.TRY_AGAIN:
				throw new UnknownHostException("Temporary failure to lookup host '" + hostname + "'");
			default:
				throw new UnknownHostException("Unknown error " + result +
						" in host lookup of '" + hostname + "'");
		}
	}

	private static final Cache dnsCache = new Cache(); // shared, thread-safe.
}
