/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client;

import java.io.IOException;
import java.io.OutputStream;
import java.net.SocketTimeoutException;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.client.connection.HttpResponseInfo;
import org.glimmerblocker.proxy.http.client.stream.TimeoutableInputStream;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.proxyRequest.ClientAbortedException;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import static dk.vitality.lib.foundation.exception.ExceptionUtils.getExMsg;

public class UrlStreamer implements Logger
{
	@NotNull protected final GlimmerGlobals globals;
	@NotNull protected final ProxyRequestHandler prh;
	@NotNull protected final HttpResponseInfo rhi;

	public UrlStreamer(@NotNull GlimmerGlobals globals,
	                   @NotNull ProxyRequestHandler prh,
	                   @NotNull HttpResponseInfo rhi)
	{
		this.globals = globals;
		this.prh = prh;
		this.rhi = rhi;
	}

	public void streamResponse(@Nullable TimeoutableInputStream is) throws Exception
	{
		prh.request.sendResponseHeaders(getResponseHeaders(), rhi.contentLength);
		if (is == null)
			return;
		final boolean debugFlag = globals.debugFlag;
		final boolean debugVerbose = debugFlag && globals.debugVerbose;
		final boolean debugHttpServer = debugFlag && globals.debugHttpServer;
		if (debugVerbose || debugHttpServer)
			ln("UrlStreamer starting, contentLength = " + rhi.contentLength);
		long st0 = System.currentTimeMillis();
		long oldestData = System.currentTimeMillis();
		OutputStream os = prh.request.getOutputStream();
		long remaining = rhi.contentLength;
		if (remaining < 0)
			remaining = Long.MAX_VALUE;
		byte[] buffer = new byte[4096];
		while (remaining > 0) {
			int num;
			long maxWait = 0;
			try {
				if (oldestData != 0) {
					long now = System.currentTimeMillis();
					maxWait = 100 - (now - oldestData);
					if (maxWait <= 0) {
						if (debugVerbose)
							ln("!!! Flushes client data, age = " + (now - oldestData) + " msecs.");
						os.flush();
						oldestData = 0;
					}
				}
				if (maxWait > 0) {
					if (debugVerbose)
						ln("Using timeout = " + maxWait + " msecs.");
					is.setReadTimeout((int)maxWait);
				} else {
					if (debugVerbose)
						ln("Using 'blocking' read.");
					is.setReadTimeout(5 * 60 * 1000);
				}
				long st1 = debugFlag ? System.currentTimeMillis() : 0;
				num = is.read(buffer);
				if (debugFlag) {
					long dur = System.currentTimeMillis() - st1;
					if (debugVerbose || dur > 500) {
						String s = "Read time: " + dur + " msecs for " + num + " bytes";
						if (rhi.contentLength >= 0)
							s += ", " + (remaining - Math.max(0, num)) + " now remaining";
						ln(s + ".");
					}
				}
				if (num > 0)
					remaining -= num;
			}
			catch (SocketTimeoutException ex) {
				if (maxWait > 0)
					continue;
				if (debugFlag)
					ln("Got exception while using 'blocking' read from server: " + getExMsg(ex));
				break;
			}
			catch (IOException ex) {
				if (debugFlag)
					ln("Got exception when reading from server: " + getExMsg(ex));
				break;
			}
			if (num < 0) {
				if (debugVerbose) {
					long dur = System.currentTimeMillis() - st0;
					ln("EOF of response data from server, total time = " + dur + " msecs.");
				}
				break;
			}
			if (num == 0)
				continue;
			if (debugVerbose)
				ln("Forwards " + num + " bytes from server");
			try {
				os.write(buffer, 0, num);
			}
			catch (IOException ex) {
				throw new ClientAbortedException("IOException: " + ex.getMessage(), ex);
			}
			if (oldestData == 0)
				oldestData = System.currentTimeMillis();
			//ln(new String(buffer, 0, num, "ISO-8859-1"));
		}
		if (debugFlag && rhi.contentLength >= 0 && remaining > 0)
			ln("Stopping with " + remaining + " bytes remaining to be read.");
		if (debugHttpServer) {
			long dur = System.currentTimeMillis() - st0;
			String s = "UrlStreamer done, duration = " + dur + " msecs";
			if (rhi.contentLength >= 0)
				s += ", " + remaining + " bytes remaining.";
			ln(s);
		}
		try {
			os.close();
		}
		catch (IOException ex) {
			throw new ClientAbortedException("IOException: " + ex.getMessage(), ex);
		}
		is.close();
		if (debugHttpServer && debugVerbose)
			ln("UrlStreamer: Virtual streams closed.");
	}

	protected Headers getResponseHeaders()
	{
		return rhi.headers;
	}

	public void ln(@Nullable String s)
	{
		prh.ln(s);
	}
}
