/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.curl;

import java.io.IOException;
import java.net.ConnectException;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.history.HistoryEntryBuilder;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.network.NetworkConfig;
import org.glimmerblocker.proxy.http.client.cache.HttpClientCache;
import org.glimmerblocker.proxy.http.client.cache.HttpRunnable;
import org.glimmerblocker.proxy.http.client.connection.HttpClientConnection;
import org.glimmerblocker.proxy.http.client.connection.HttpResponseInfo;
import org.glimmerblocker.proxy.http.client.connection.HttpTransaction;
import org.glimmerblocker.proxy.http.client.stream.RequestBody;
import org.glimmerblocker.proxy.http.client.stream.RequestData;
import org.glimmerblocker.proxy.http.client.stream.TimeoutableInputStream;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.rule.AbstractRule;
import org.glimmerblocker.proxy.rule.RequestRule;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import static java.net.HttpURLConnection.HTTP_BAD_GATEWAY;
import static java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT;
import static java.net.HttpURLConnection.HTTP_NOT_FOUND;

/**
 * Base-class for a 'curl'-like class to get content from an HTTP server.
 * Sub-classes streams the content to the client or saves it in a byte array.
 * <p/>
 * It currently uses {@link java.net.HttpURLConnection} which isn't optimal.
 */
public abstract class CurlBase implements Logger
{
	@NotNull public final GlimmerGlobals globals;
	/**
	 * The HTTP method, e.g. GET.
	 */
	@NotNull public String method;

	/**
	 * The original requested URL.
	 */
	@NotNull public final URL originalUrl;

	/**
	 * Final URL after handling redirects
	 */
	@NotNull public URL url;

	/**
	 * The request headers, e.g. "User-Agent".
	 */
	@NotNull public Headers requestHeaders;

	/**
	 * Network configuration: http proxy, socks proxy, dns server.
	 */
	@NotNull public NetworkConfig networkConfig;

	/**
	 * Should redirects from the server be handled automatically, or do we want to return the redirect
	 * headers to the client.
	 */
	public boolean followRedirects = false;

	/**
	 * The HTTP headers sent by the server.
	 */
	public Headers responseHeaders;

	/**
	 * Number of redirects handled to avoid infinite loops from the server.
	 */
	protected int redirectCount;

	/**
	 * The proxy requests logger.
	 */
	@NotNull private final Logger parentLogger;

	/**
	 * Http client addr (if any).
	 */
	@Nullable public final InetAddress clientAddr;

	protected CurlBase(@NotNull GlimmerGlobals globals,
	                   @NotNull Logger parentLogger,
	                   @NotNull NetworkConfig networkConfig,
	                   @Nullable InetAddress clientAddr,
	                   @NotNull String method,
	                   @NotNull URL url,
	                   @NotNull Headers requestHeaders) throws Exception
	{
		this.globals = globals;
		this.parentLogger = parentLogger;
		this.networkConfig = networkConfig;
		this.clientAddr = clientAddr;
		this.method = method;
		this.originalUrl = url;
		this.url = url;
		this.requestHeaders = requestHeaders;
		//noinspection ConstantConditions
		if (url == null)
			throw new NullPointerException("this.url is null");
	}

	/**
	 * Main method for executing the request.
	 * Autmatically handles redirects if {@link #followRedirects} is true.
	 */
	public void execute() throws Exception
	{
		for (redirectCount = 0; true; redirectCount++) {
			checkRedirectCount();
			if (processRequestRules())
				return;
			executeOneRequest();
			if (responseHeaders == null) {
				// this can happend in UnmodifiedStreamer when a UnknownHostException is thrown.
				return;
			}
			if (!isRedirect() || !followRedirects || !handleRedirect())
				return;
		}
	}

	protected void executeOneRequest() throws Exception
	{
		try {
			executeWithoutExceptionHandling(globals.httpClientCache);
		}
		catch (ConnectException ex) {
			Headers h = new Headers(HTTP_GATEWAY_TIMEOUT, "Server denies connection.");
			throw new HttpException(h, ex);
		}
		catch (SocketTimeoutException ex) {
			Headers h = new Headers(HTTP_GATEWAY_TIMEOUT, "Server doesn't respond at all.");
			throw new HttpException(h, ex);
		}
		catch (UnknownHostException ex) {
			Headers h = new Headers(HTTP_NOT_FOUND, "Host not found: " + url);
			throw new HttpException(h, ex);
		}
		catch (IOException ex) {
			Headers h = new Headers(HTTP_BAD_GATEWAY, "Got " + ex.getClass().getName() +
					" (" + ex.getMessage() + "): " + url);
			if (globals.debugHttpClient.get())
				ex.printStackTrace();
			throw new HttpException(h, ex);
		}
	}

	protected void executeWithoutExceptionHandling(HttpClientCache clientCache) throws Exception
	{
		//noinspection ConstantConditions
		if (url == null)
			throw new NullPointerException("this.url is null");
		String protocol = url.getProtocol();
		int port = url.getPort();
		if (port <= 0)
			port = protocol.equals("https") ? 443 : 80;
		String host = url.getHost();
		clientCache.execute(networkConfig, protocol, host, port, createHttpRunnable());
	}

	protected HttpRunnable createHttpRunnable() throws Exception
	{
		return new HttpRunnableBase();
	}

	protected class HttpRunnableBase implements HttpRunnable
	{
		public void run(@NotNull HttpClientConnection conn) throws Exception
		{
			HttpTransaction t = makeTransaction(conn);
			t.sendRequest();
			HttpResponseInfo hri = t.getResponseHeaderInfo();
			responseHeaders = hri.headers;
			if (followRedirects && isRedirect())
				return;
			HttpResponseInfo c = t.getOpt100ContinueResponseHeaderInfo();
			if (c != null)
				handle100ContinueResponse(c);
			TimeoutableInputStream is = hri.isContentLessResponse ? null : t.getBodyStream();
			readBody(hri, is);
		}

		protected HttpTransaction makeTransaction(HttpClientConnection conn) throws Exception
		{
			RequestData rd = makeRequestData();
			return conn.makeTransaction(rd);
		}
	}

	protected RequestData makeRequestData() throws Exception
	{
		return new RequestData(method, url, requestHeaders, getRequestBody());
	}

	protected RequestBody getRequestBody() throws Exception
	{
		return new RequestBody(null, -1);
	}

	/**
	 * @param bodyStream is null if the response is body-less (e.g. HEAD or 304 NotModified)
	 */
	protected abstract void readBody(@NotNull HttpResponseInfo hri,
	                                 @Nullable TimeoutableInputStream bodyStream) throws Exception;


	protected abstract void sendRequestRuleData(@NotNull ResultData data) throws Exception;

	@SuppressWarnings({"UnusedDeclaration"})
	protected void handle100ContinueResponse(@NotNull HttpResponseInfo hri) throws Exception
	{
		// ignored.
	}

	/**
	 * See {@link Logger#ln}
	 */
	public void ln(@Nullable String s)
	{
		parentLogger.ln(s);
	}

	protected boolean processRequestRules() throws Exception
	{
		String host = url.getHost();
		String requestPath = url.getPath();
		if (requestPath == null)
			requestPath = "/";
		String query = url.getQuery();
		if (query == null)
			query = "";
		boolean mobile = isMobileUserAgent();
		List<AbstractRule> ruleList = globals.ruleCollection.getRules(host, requestPath, query, mobile);
		if (ruleList.isEmpty())
			return false;
		List<RequestRule> requestList = null;
		for (AbstractRule rule : ruleList) {
			if (!(rule instanceof RequestRule))
				continue;
			if (requestList == null)
				requestList = new ArrayList<>(ruleList.size());
			requestList.add((RequestRule)rule);
		}
		return (requestList != null) && processRequestRules(requestList);
	}

	protected boolean processRequestRules(@NotNull List<RequestRule> ruleList) throws Exception
	{
		@Nullable HistoryEntryBuilder heb = getHistoryEntryBuilderForRequestRules();
		for (RequestRule rule : ruleList) {
			if (heb != null)
				heb.addAction(rule, "request");
			RequestScriptExecutor exec = new RequestScriptExecutor(this, clientAddr, rule);
			if (exec.rd != null) {
				sendRequestRuleData(exec.rd);
				return true;
			}
		}
		return false;
	}

	@Nullable
	protected HistoryEntryBuilder getHistoryEntryBuilderForRequestRules()
	{
		return null;
	}

	protected boolean isMobileUserAgent()
	{
		return requestHeaders.hasMobileUserAgent();
	}

	/**
	 * Checks the number of redirects to avoid infinite loops from the server.
	 */
	protected void checkRedirectCount() throws Exception
	{
		if (redirectCount < 25)
			return;
		Headers h = new Headers(HTTP_BAD_GATEWAY, "Too many redirects from the server.",
				"text/plain", null, null);
		throw new HttpException(h);
	}

	/**
	 * Does the HTTP status from the server specify a redirect status (301, 302, 303).
	 */
	protected boolean isRedirect()
	{
		switch (responseHeaders.status) {
			case HttpURLConnection.HTTP_MOVED_PERM:
			case HttpURLConnection.HTTP_MOVED_TEMP:
			case HttpURLConnection.HTTP_SEE_OTHER:
				return true;
			default:
				return false;
		}
	}

	/**
	 * Handle a redirect, assuming both {@link #followRedirects} and {@link #isRedirect} are true.
	 * Updates {@link #url} with the new URL.
	 */
	protected boolean handleRedirect() throws Exception
	{
		String s = responseHeaders.headerMap.get("location").get(0);
		if (s == null)
			throw new Exception("Redirect without 'location' header");
		if (globals.debugFlag)
			ln("Redirect to: " + s);
		url = new URL(url, s);
		if (globals.debugFlag && !s.contains("://"))
			ln("   absolute: " + url);
		if (responseHeaders.status == HttpURLConnection.HTTP_SEE_OTHER)
			method = "GET"; // rfc 2616, section 10.3.4
		return true;
	}
}
