/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.data;

import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.proxyRequest.transform.MimeTypeMatcher;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import static java.util.regex.Pattern.CASE_INSENSITIVE;

/**
 * http://www.whatwg.org/specs/web-apps/current-work/multipage/parsing.html
 */
public class CharsetDeterminer implements Logger
{
	@NotNull public final ResultData rd;
	@NotNull public final String httpCharset;
	@NotNull public final String javaCharset;

	@NotNull private final Logger parentLogger;
	private final boolean debugFlag;

	private int bomSize;
	private boolean hasCharsetInHttpHeader;

	public CharsetDeterminer(@NotNull ResultData rd,
	                         @NotNull Logger logger,
	                         boolean debugFlag) throws Exception
	{
		this.rd = rd;
		this.parentLogger = debugFlag ? new PrefixedLogger(logger, "CharsetDecoder") : logger;
		this.debugFlag = debugFlag;
		this.httpCharset = findCharset();
		this.javaCharset = (httpCharset.equalsIgnoreCase("gb2312")) ? "ISO-8859-1" : httpCharset;
		/*
			The reason for decoding GB2312 as ISO-8859-1 is:
			 1) Many pages sent as GB2312 uses a superset charset of GB2312, and using GB2312
			    results in invalid text, just like using strict "ASCII" fails to decode and recode UTF-8 input
			    (as all 8-bit chars would have been translated into e.g. "?").
			 2) ISO-8859-1 decodes the bytes 1:1 into chars and thus is transparent when recoding to bytes.
			 3) The only drawback is that the chars in Javascript are invalid, so you cannot do e.g.
			    replace(/鮋妦𢏐猁蚚/, "something else");
			    as "鮋妦𢏐猁蚚" is represented as approx. "éÊ²üNÒªÓÃ" in the Javascript 't' variable.
		 */
	}

	@NotNull
	public String getContentAsText() throws Exception
	{
		if (debugFlag)
			ln("Decodes using charset '" + httpCharset + "'");
		byte[] b = rd.getContent();
		if (b == null)
			return "";
		return new String(b, bomSize, b.length - bomSize, javaCharset);
	}

	public boolean hasCharsetInHttpHeader()
	{
		return hasCharsetInHttpHeader;
	}

	private String findCharset() throws Exception
	{
		String cs = checkHttpHeader();
		hasCharsetInHttpHeader = (cs != null);
		if (cs == null)
			cs = checkBOM();
		else
			checkBOM(); // find bomSize
		// Use status = OK for all cases as we want to be able to sniff html even for "500 internal error".
		if (cs == null && MimeTypeMatcher.Html.matches(HttpURLConnection.HTTP_OK, rd.headers.contentTypeBaseLC))
			cs = sniffHtml();
		if (cs == null) {
			cs = "ISO-8859-1";
			if (debugFlag)
				ln("No charset found at all, using default.");
		}
		return cs;
	}

	@Nullable
	private String checkHttpHeader()
	{
		final String ct = rd.headers.contentType;
		if (ct == null)
			return null;
		Matcher m = quotedCharsetRE.matcher(ct);
		if (m.find()) {
			String s = m.group(2);
			if (debugFlag)
				ln("Found quoted charset in content-type http-header.");
			return s;
		}
		m = unquotedCharsetRE.matcher(ct);
		if (m.find()) {
			String s = m.group(1);
			if (debugFlag)
				ln("Found unquoted charset in content-type http-header.");
			return s;
		}
		return null;
	}

	@Nullable
	private String checkBOM() throws UnsupportedEncodingException
	{
		byte[] b = rd.getContent();
		if (b == null)
			return null;
		if (b.length < 2)
			return null;
		int b0 = 255 & b[0];
		int b1 = 255 & b[1];
		if (b.length >= 2 && b0 == 0xFE && b1 == 0xFF) {
			bomSize = 2;
			if (debugFlag)
				ln("Found UTF-16BE BOM");
			return "UTF-16BE";
		}
		if (b.length >= 2 && b0 == 0xFF && b1 == 0xFE) {
			bomSize = 2;
			if (debugFlag)
				ln("Found UTF-16LE BOM");
			return "UTF-16LE";
		}
		if (b.length < 3)
			return null;
		int b2 = b[2];
		if (b.length >= 3 && b0 == 0xEF && b1 == 0xBB && b2 == 0xBF) {
			bomSize = 3;
			if (debugFlag)
				ln("Found UTF-8 BOM");
			return "UTF-8";
		}
		return null;
	}

	private String sniffHtml() throws Exception
	{
		Parser p = new Parser(this);
		return p.charset;
	}

	public void ln(@Nullable String s)
	{
		parentLogger.ln(s);
	}

	private static class Parser implements Logger
	{
		private final byte[] b;
		private final int maxIdx;
		private final boolean debugFlag;
		private final CharsetDeterminer determiner;
		private int idx;
		String charset;
		//
		private String attrName, attrValue;
		private StringBuilder sb = new StringBuilder(50);

		Parser(CharsetDeterminer determiner) throws Exception
		{
			b = determiner.rd.getContent();
			this.debugFlag = determiner.debugFlag;
			this.determiner = determiner;
			if (b == null) {
				maxIdx = -1;
				return;
			}
			//
			maxIdx = b.length - 16; // 16 allows for dropping array index checks
			while (idx < maxIdx) {
				byte ch = b[idx++];
				if (ch != '<')
					continue;
				if (has("!--")) {
					int x = idx - 1;
					idx += 3;
					while (idx < maxIdx && !has("-->"))
						idx++;
					idx += 3;
					if (debugFlag)
						ln("Skipped comment: {{{" + new String(b, x, idx - x, "ISO-8859-1") + "}}}");
					continue;
				}
				if (has("meta") && isWhitespace(b[idx + 4])) {
					int x = idx - 1;
					idx += 4;
					String content = null;
					boolean isContentType = false;
					while (parseAttribute()) {
						if (attrName.equals("charset")) {
							charset = attrValue;
							if (debugFlag)
								ln("Found <meta charset='" + attrValue + "'>");
							return;
						}
						if (attrName.equals("content"))
							content = attrValue;
						else if (attrName.equals("http-equiv") && attrValue.equals("content-type"))
							isContentType = true;
					}
					if (isContentType && content != null) {
						if (debugFlag)
							ln("<mete http-equiv=ct:  value='" + content + "'");
						Matcher m = unquotedCharsetRE.matcher(content);
						if (m.find()) {
							charset = m.group(1);
							if (debugFlag)
								ln("Found <meta content='...charset=" + attrValue + "'...>");
							return;
						}
						if (debugFlag) {
							ln("Could not match charset in <meta>:  {{{" +
									new String(b, x, idx - x, "ISO-8859-1") + "}}}");
						}
					}
					continue;
				}
				if (has("body") && !isAsciiLetter(b[idx + 4]))
					return;
				ch = b[idx++];
				if (isAsciiLetter(ch)) {
					skipTag();
					continue;
				}
				if (ch == '/') {
					if (has("head") && !isAsciiLetter(b[idx + 4]))
						return;
					ch = b[idx++];
					if (isAsciiLetter(ch))
						skipTag();
					else
						skipPastGt();
					continue;
				}
				if (ch == '!' || ch == '?')
					skipPastGt();
			}
		}

		private boolean has(String s)
		{
			int len = s.length();
			if (idx + len > maxIdx)
				return false;
			for (int i = 0; i < len; i++) {
				if (Character.toLowerCase(b[idx + i]) != s.charAt(i))
					return false;
			}
			return true;
		}

		private void skipPastGt()
		{
			while (idx < maxIdx && b[idx] != '>')
				idx++;
			idx++;
		}

		private boolean isWhitespace(byte ch)
		{
			return ch == 0x09 || ch == 0x0A || ch == 0x0C || ch == 0x0D || ch == 0x20;
		}

		private boolean isAsciiLetter(byte ch)
		{
			return (ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z');
		}

		private void skipWhitespace()
		{
			while (idx <= maxIdx && isWhitespace(b[idx]))
				idx++;
		}

		private void skipTag() throws Exception
		{
			int x = idx - 1;
			while (idx < maxIdx) {
				byte ch = b[idx++];
				if (ch == '>') {
					if (debugFlag)
						ln("Skipped no-attr tag:  {{{" + new String(b, x, idx - x, "ISO-8859-1") + "}}}");
					return;
				}
				if (isWhitespace(ch))
					break;
			}
			while (parseAttribute()) {
				if (debugFlag)
					ln("Got attr: {{{" + attrName + "}}} = {{{" + attrValue + "}}}");
			}
			if (debugFlag)
				ln("Skipped attr tag:  {{{" + new String(b, x, idx - x, "ISO-8859-1") + "}}}");
		}

		private boolean parseAttribute()
		{
			attrName = "";
			attrValue = "";
			skipWhitespace();
			sb.setLength(0);
			while (true) {
				if (idx > maxIdx)
					return false; // name doesn't matter as attrValue must be non-empty for charsets
				byte ch = b[idx++];
				if (ch == '>')
					return false;
				if (ch == '=' && sb.length() > 0)
					break;
				if (isWhitespace(ch)) {
					skipWhitespace();
					if (b[idx] == '=') {
						idx++;
						break;
					}
					return false; // empty attr.
				}
				sb.append(Character.toLowerCase((char)ch));
			}
			// here: just past '='
			attrName = sb.toString();
			sb.setLength(0);
			skipWhitespace();
			byte ch = b[idx++];
			if (ch == '>')
				return false; // empty attr.
			if (ch != '"' && ch != '\'') {
				sb.append(Character.toLowerCase((char)ch));
				while (idx <= maxIdx) {
					ch = b[idx];
					if (ch == '>') // need reparse this byte to end element, but must return true this time.
						break;
					idx++;
					if (isWhitespace(ch))
						break;
					sb.append(Character.toLowerCase((char)ch));
				}
				attrValue = sb.toString();
				return true;
			}
			byte q = ch;
			while (idx <= maxIdx) {
				ch = b[idx++];
				if (ch == q)
					break;
				sb.append(Character.toLowerCase((char)ch));
			}
			attrValue = sb.toString();
			return true;
		}

		public void ln(@Nullable String s)
		{
			determiner.ln(s);
		}
	}

	private static final Pattern quotedCharsetRE =
			Pattern.compile("\\bcharset\\s*=\\s*(['\"])([^<>]+)\\1", CASE_INSENSITIVE);

	private static final Pattern unquotedCharsetRE =
			Pattern.compile("\\bcharset\\s*=\\s*([-_A-Za-z0-9]+)", CASE_INSENSITIVE);
}
