/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.server;

import java.io.EOFException;
import java.io.IOException;
import java.io.PushbackInputStream;
import javax.net.ssl.HttpsURLConnection;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.client.connection.KeepAliveException;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.misc.ByteBuffer;

class HttpRequestHeaderReader implements Logger
{
	@NotNull private final HttpServerConnection conn;
	@NotNull private final HttpServerRequest req;
	@NotNull private final PushbackInputStream pbis;
	final boolean debugFlag;

	String method, uri;
	Headers requestHeaders;
	boolean http11;
	long contentLength = -1;
	String transferEncoding;

	private String name;
	@NotNull private final ByteBuffer bb = new ByteBuffer(50);
	private int headerSize;

	HttpRequestHeaderReader(@NotNull HttpServerRequest req) throws Exception
	{
		this.conn = req.conn;
		this.req = req;
		this.debugFlag = req.globals.debugHttpServer;
		//
		conn.skt.setSoTimeout(60 * 1000);
		this.pbis = new PushbackInputStream(conn.is, 1);
		parseRequestLine(readRequestLine());
		HeadersBuilder hb = HeadersBuilder.makeEmpty();
		String connectionOptions = null;
		while (true) {
			name = null;
			readName();
			if (name == null)
				break;
			String value = readLine(true);
			if (debugFlag)
				ln("  " + name + ": " + value);
			String nameLC = name.toLowerCase();
			switch (nameLC) {
				case "content-encoding":
					hb.contentEncoding = value;
					break;
				case "content-type":
					hb.contentType = value;
					break;
				case "content-length":
					String s = zapOptions(value);
					try {
						contentLength = Long.parseLong(s);
					}
					catch (Exception ex) {
						ln("Invalid content-length: '" + s + "': " + ExceptionUtils.getExMsg(ex));
						Headers h = new Headers(HttpsURLConnection.HTTP_BAD_REQUEST, "Invalid ContentLength header");
						throw new HttpException(h, ex);
					}
					break;
				case "transfer-encoding":
					if (value.equalsIgnoreCase("identity"))
						transferEncoding = null; // explicit value is just noise as identity is the default.
					else
						transferEncoding = value;
					break;
				case "proxy-connection":
					if (hasCloseOption(value))
						conn.disableKeepalive("Got 'Proxy-Connection: " + value + "'");
					break;
				case "connection":
					connectionOptions = value;
					break;
				default:
					hb.add(nameLC, value);
					break;
			}
		}
		if (connectionOptions != null && hasCloseOption(connectionOptions)) {
			// even if we have 'proxy-connection', play safe and adhere a 'close' in the 'connection' header
			conn.disableKeepalive("Got 'Connection: " + connectionOptions + "'");
		}
		if (transferEncoding != null && contentLength >= 0) {
			if (debugFlag) {
				ln("Request has specified both Content-Length and Transfer-Encoding which is illegal." +
						" Ignores Content-Length.");
			}
			contentLength = -1;
		}
		requestHeaders = hb.toHeaders();
		if (debugFlag)
			ln("Done reading request header.");
	}

	private String readRequestLine() throws Exception
	{
		String requestLine;
		try {
			// some servers send too many LF in keep-alive connections.
			int i = 0;
			do {
				requestLine = readLine(false);
			} while (requestLine.length() == 0 && ++i < 5);
		}
		catch (IOException ex) {
			String msg = "Client closed connection before sending request #" + conn.requestCount;
			throw new KeepAliveException(msg, ex);
		}
		if (debugFlag)
			ln("Request line: '" + requestLine + "'");
		return requestLine;
	}

	private void parseRequestLine(String requestLine) throws Exception
	{
		RequestLineTokenizer t = new RequestLineTokenizer(requestLine);
		method = t.getNextToken();
		uri = t.getNextToken();
		String v = t.getNextToken();
		if (method.length() == 0)
			throwInvalidRequestLine(requestLine, "Empty http method");
		if (uri.length() == 0)
			throwInvalidRequestLine(requestLine, "Empty URI path");
		if (t.getNextToken().length() > 0)
			throwInvalidRequestLine(requestLine, "More than 3 tokens");
		http11 = false;
		if (v.length() == 0)
			conn.disableKeepalive("HTTP/0.9");
		else if (v.equals("HTTP/1.0"))
			conn.disableKeepalive(v);
		else
			http11 = true;
	}

	private void throwInvalidRequestLine(String requestLine, String why) throws HttpException
	{
		String msg = "Invalid request line:" + why + ", line is '" + requestLine + "'";
		throw new HttpException(new Headers(HttpsURLConnection.HTTP_BAD_REQUEST, msg));
	}

	public void ln(@Nullable String s)
	{
		req.ln(s);
	}

	private boolean hasCloseOption(String s)
	{
		return s.equalsIgnoreCase("close") || s.toLowerCase().contains("close");
	}

	private String zapOptions(String s)
	{
		int idx = s.indexOf(';');
		return (idx < 0) ? s.trim() : s.substring(0, idx).trim();
	}

	private void readName() throws Exception
	{
		bb.clear();
		int ch;
		do {
			ch = read();
		} while (ch == 32);
		if (ch == 13)
			ch = read();
		if (ch == 10) {
			name = null;
			return;
		}
		while (ch != ':' && ch != 13 && ch != 10) {
			if (ch != ' ')
				bb.append(ch);
			ch = read();
		}
		if (ch != ':') {
			throw new IOException("Invalid http header: name '"
					+ bb.toString("ISO-8859-1") + "' not terminated by ':'");
		}
		if (bb.length() == 0)
			throw new IOException("Invalid http header: empty name");
		name = decode();
	}

	String readLine(boolean allowContinuations) throws Exception
	{
		bb.clear();
		while (true) {
			int ch = read();
			if (ch < 0)
				break;
			if (allowContinuations && (ch == 32 || ch == '\t') && bb.length() == 0)
				continue; // space after colon.
			if (ch == 13)
				continue;
			if (ch == 10) {
				if (allowContinuations) {
					int ch2 = pbis.read();
					if (ch2 == 32 || ch2 == '\t') {
						// folded header line, rfc 2616, 4.2
						continue;
					}
					pbis.unread(ch2);
				}
				break;
			}
			if (ch < 32)
				ch = 32;
			bb.append(ch);
		}
		return decode();
	}

	private int read() throws IOException
	{
		int ch = pbis.read();
		if (ch < 0)
			throw new EOFException("Client closed connection while sending the HTTP request header.");
		if (headerSize++ >= 65536)
			throw new IOException("Too huge HTTP header (64k)");
		return (char)ch;
	}

	private String decode() throws Exception
	{
		//ln("Decode: '" + bb.toString("ISO-8859-1") + "' from " + HexUtils.toLcHexSpaced(bb.toBytes()));
		return bb.toString("ISO-8859-1");
	}

	static class RequestLineTokenizer
	{
		@NotNull final String line;
		int idx;

		RequestLineTokenizer(@NotNull String line)
		{
			this.line = line;
		}

		String getNextToken()
		{
			int len = line.length();
			while (idx < len && line.charAt(idx) <= 32)
				idx++;
			int start = idx;
			while (idx < len && line.charAt(idx) > 32)
				idx++;
			return idx > start ? line.substring(start, idx) : "";
		}
	}
}
