/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.proxyRequest.keyword;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.network.NetworkConfig;
import org.glimmerblocker.proxy.http.client.curl.RequestGlobalObject;
import org.glimmerblocker.proxy.http.client.curl.RequestSubObject;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.glimmerblocker.proxy.rhino.callback.CallbackObjectBase;
import org.glimmerblocker.proxy.rhino.callback.CallbackRhinoBinder;
import org.glimmerblocker.proxy.rhino.callback.RhinoCallback;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.Scriptable;
import dk.vitality.lib.foundation.html.URLUtils;
import dk.vitality.lib.foundation.io.SimpleUnixExec;
import dk.vitality.lib.foundation.ref.FatRef;

public class KeywordGlobalObject extends RequestGlobalObject
{
	ProxyRequestHandler prh;
	String keyword;
	String term, encodedTerm;

	private static final CallbackRhinoBinder classBinder = new CallbackRhinoBinder(KeywordGlobalObject.class);

	public KeywordGlobalObject(@NotNull GlimmerGlobals globals,
	                           @NotNull Scriptable scope,
	                           @NotNull Scriptable prototype) throws IOException
	{
		super(globals, scope, prototype, classBinder);
	}

	public void prepareKeywordCallbacks(@NotNull Logger logger,
	                                    @NotNull Context ctx,
	                                    @NotNull ProxyRequestHandler prh,
	                                    @NotNull String keyword,
	                                    @NotNull String term) throws Exception
	{
		super.prepareRequestObject(ctx, logger, prh.request.getClientAddress(),
				prh.request.method, prh.requestURL, prh.request.headers,
				new FatRef<>(prh.networkConfig));
		this.prh = prh;
		this.keyword = keyword;
		this.term = term;
		this.encodedTerm = URLUtils.encodeURL(term);
	}

	@Override
	public void notifyLeftContext()
	{
		super.notifyLeftContext();
		prh = null;
		keyword = null;
		term = null;
		encodedTerm = null;
	}

	//===========================================================================

	protected RequestSubObject createRequestSubObject() throws Exception
	{
		return new KeywordRequestSubObject(this);
	}

	private static final CallbackRhinoBinder reqBinder = new CallbackRhinoBinder(KeywordRequestSubObject.class);

	class KeywordRequestSubObject extends RequestSubObject
	{
		KeywordRequestSubObject(@NotNull CallbackObjectBase parent)
		{
			super(parent, reqBinder);
		}

		@RhinoCallback
		public String jsGet_keyword()
		{
			return keyword;
		}

		@RhinoCallback
		public String jsGet_term()
		{
			return term;
		}

		@RhinoCallback
		public String jsGet_encodedTerm()
		{
			return encodedTerm;
		}
	}
	//===========================================================================

	@RhinoCallback
	public String jsGet_keyword()
	{
		return keyword;
	}

	@RhinoCallback
	public String jsGet_term()
	{
		return term;
	}

	@RhinoCallback
	public String jsGet_encodedTerm()
	{
		return encodedTerm;
	}

	@Nullable
	@RhinoCallback
	public String jsGet_url() throws Exception
	{
		return jsGet_response().jsGet_headers().getHeaderLC("location");
	}


	@RhinoCallback
	public void jsSet_url(String url) throws Exception
	{
		jsGet_response().jsFunction_sendRedirect(url);
	}

	@RhinoCallback
	public String jsFunction_unixManPage(String name, String section) throws Exception
	{
		if (name == null)
			throw new IOException("name not specified");
		if (!isSafeParam(name))
			throw new IOException("Illegal command name.");
		if (section != null && section.length() > 0) {
			if (!isSafeParam(section))
				throw new IOException("Illegal section name.");
		} else {
			section = null;
		}
		try {
			String[] cmd = (section != null)
					? new String[]{"man", section, name}
					: new String[]{"man", name};
			if (globals.debugFlag)
				ln(Arrays.toString(cmd));
			SimpleUnixExec sue = new SimpleUnixExec(cmd);
			sue.setWantsStdoutAsBytes(true);
			sue.setWorkingDir(new File("/Library/GlimmerBlocker"));
			sue.run();
			if (sue.getResultCode() != 0) {
				if (globals.debugFlag)
					ln("man exit code = " + sue.getResultCode());
				return null;
			}
			byte[] b = sue.getStdoutAsBytes();
			//
			String k = keyword.equals("m") ? "m" : "man";
			String[] cmd2 = {
					"perl",
					"-T",
					new File(GlimmerGlobals.proxyResourcesF.getPath(), "man2html").getPath(),
					"--cgiurlexp=\"http://" + k + "%20$section%20$title\";"};
			if (globals.debugFlag)
				ln(Arrays.toString(cmd2));
			sue = new SimpleUnixExec(cmd2);
			sue.setWantsStdoutAsString(true);
			sue.setWorkingDir(new File("/Library/GlimmerBlocker"));
			sue.setStdin(b);
			sue.run();
			if (sue.getResultCode() != 0) {
				ln("man2html failed (" + sue.getResultCode() + " formatting man page: " + name);
				return null;
			}
			return sue.getStdoutAsString();
		}
		finally {
			zap("/Library/GlimmerBlocker/.lesshst");
			zap("/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/Resources/.lesshst");
		}
	}

	private void zap(String path)
	{
		File f = new File(path);
		if (!f.exists())
			return;
		if (!f.delete())
			ln("Can't delete " + f);
		else if (globals.debugFlag)
			ln("Did delete " + f);
	}

	private boolean isSafeParam(String s)
	{
		if (s == null)
			return true;
		int len = s.length();
		if (len > 200)
			return false;
		if (s.startsWith("/") || s.contains("../"))
			return false;
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch < 32 || ch > 255)
				return false;
			if (ch >= 127 && ch <= 160) // undefined in latin-1
				return false;
		}
		return true;
	}
}
